"""Tests for tmodbus/pdu/base.py ."""

import pytest
from tmodbus.exceptions import InvalidRequestError, InvalidResponseError
from tmodbus.pdu.base import (
    BaseClientPDU,
    BasePDU,
    BaseSubFunctionClientPDU,
    BaseSubFunctionPDU,
)


class TestBaseClientPDU:
    """Tests for BaseClientPDU."""

    def test_get_expected_response_data_length_with_fixed_length(self) -> None:
        """Test get_expected_response_data_length with rtu_response_data_length set."""

        class TestPDU(BaseClientPDU[int]):
            function_code = 0x01
            rtu_response_data_length = 10

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        assert TestPDU.get_expected_response_data_length(b"") == 10

    def test_get_expected_response_data_length_from_first_byte(self) -> None:
        """Test get_expected_response_data_length when length is in first byte."""

        class TestPDU(BaseClientPDU[int]):
            function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # First byte is 0x0f (15), so expected length is 1 + 15 = 16
        assert TestPDU.get_expected_response_data_length(b"\x0f") == 16
        # First byte is 0x05 (5), so expected length is 1 + 5 = 6
        assert TestPDU.get_expected_response_data_length(b"\x05") == 6


class TestBasePDU:
    """Tests for BasePDU."""

    def test_get_expected_response_data_length_with_fixed_length(self) -> None:
        """Test expected response data length for BasePDU with fixed length."""

        class TestPDU(BasePDU[int]):
            function_code = 0x01
            rtu_response_data_length = 10

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        assert TestPDU.get_expected_response_data_length(b"") == 10

    def test_get_expected_response_data_length_from_first_byte(self) -> None:
        """Test expected response data length when length is in first byte."""

        class TestPDU(BasePDU[int]):
            function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        assert TestPDU.get_expected_response_data_length(b"\x0f") == 16

    def test_get_expected_request_data_length_with_fixed_length(self) -> None:
        """Test get_expected_request_data_length with rtu_request_data_length set."""

        class TestPDU(BasePDU[int]):
            function_code = 0x01
            rtu_request_data_length = 8

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Should return the fixed length
        assert TestPDU.get_expected_request_data_length(b"") == 8

    def test_get_expected_request_data_length_from_first_byte(self) -> None:
        """Test get_expected_request_data_length when length is in first byte."""

        class TestPDU(BasePDU[int]):
            function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # First byte is 0x0a (10), so expected length is 1 + 10 = 11
        assert TestPDU.get_expected_request_data_length(b"\x0a") == 11
        # First byte is 0x03 (3), so expected length is 1 + 3 = 4
        assert TestPDU.get_expected_request_data_length(b"\x03") == 4


class TestBaseSubFunctionClientPDU:
    """Tests for BaseSubFunctionClientPDU."""

    def test_get_expected_response_data_length_invalid_sub_function_code(self) -> None:
        """Test that invalid sub-function code raises InvalidResponseError."""

        class TestPDU(BaseSubFunctionClientPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Data with wrong sub-function code (0x0F instead of 0x0E)
        with pytest.raises(InvalidResponseError) as exc_info:
            TestPDU.get_expected_response_data_length(b"\x0f\x10")

        assert "Expected sub-function code 14, got 15" in str(exc_info.value)

    def test_get_expected_response_data_length_with_fixed_length(self) -> None:
        """Test get_expected_response_data_length with rtu_response_data_length set."""

        class TestPDU(BaseSubFunctionClientPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E
            rtu_response_data_length = 20

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Should return the fixed length when sub-function code matches
        assert TestPDU.get_expected_response_data_length(b"\x0e\x00") == 20

    def test_get_expected_response_data_length_from_second_byte(self) -> None:
        """Test get_expected_response_data_length when length is in second byte."""

        class TestPDU(BaseSubFunctionClientPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # First byte is sub-function (0x0E), second byte is length (0x0C = 12)
        # Expected: 1 (sub-function) + 1 (length byte) + 12 = 14
        assert TestPDU.get_expected_response_data_length(b"\x0e\x0c") == 14
        # First byte is sub-function (0x0E), second byte is length (0x05 = 5)
        # Expected: 1 + 1 + 5 = 7                               # noqa: ERA001
        assert TestPDU.get_expected_response_data_length(b"\x0e\x05") == 7


class TestBaseSubFunctionPDU:
    """Tests for BaseSubFunctionPDU."""

    def test_get_expected_response_data_length_invalid_sub_function_code(self) -> None:
        """Test that invalid sub-function code raises InvalidResponseError."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Data with wrong sub-function code
        with pytest.raises(InvalidResponseError) as exc_info:
            TestPDU.get_expected_response_data_length(b"\x0a\x10")

        assert "Expected sub-function code 14, got 10" in str(exc_info.value)

    def test_get_expected_response_data_length_with_fixed_length(self) -> None:
        """Test get_expected_response_data_length with rtu_response_data_length set."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E
            rtu_response_data_length = 25

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Should return the fixed length
        assert TestPDU.get_expected_response_data_length(b"\x0e\x00") == 25

    def test_get_expected_response_data_length_from_second_byte(self) -> None:
        """Test get_expected_response_data_length when length is in second byte."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Second byte contains length
        assert TestPDU.get_expected_response_data_length(b"\x0e\x08") == 10
        assert TestPDU.get_expected_response_data_length(b"\x0e\x14") == 22

    def test_get_expected_request_data_length_invalid_sub_function_code(self) -> None:
        """Test that invalid sub-function code raises InvalidRequestError."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Data with wrong sub-function code (0x0D instead of 0x0E)
        with pytest.raises(InvalidRequestError) as exc_info:
            TestPDU.get_expected_request_data_length(b"\x0d\x10")

        assert "Expected sub-function code 14, got 13" in str(exc_info.value)

    def test_get_expected_request_data_length_with_fixed_length(self) -> None:
        """Test get_expected_request_data_length with rtu_request_data_length set."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E
            rtu_request_data_length = 15

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # Should return the fixed length when sub-function code matches
        assert TestPDU.get_expected_request_data_length(b"\x0e\x00") == 15

    def test_get_expected_request_data_length_from_first_byte(self) -> None:
        """Test get_expected_request_data_length when length is in first byte."""

        class TestPDU(BaseSubFunctionPDU[int]):
            function_code = 0x2B
            sub_function_code = 0x0E

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

            @classmethod
            def decode_request(cls, _request: bytes) -> "TestPDU":
                return cls()

            def encode_response(self, _value: int) -> bytes:
                return b""

        # First byte is sub-function code (0x0E = 14)
        # Expected: 1 (length byte) + 14 = 15
        assert TestPDU.get_expected_request_data_length(b"\x0e") == 15
        # First byte is different value
        with pytest.raises(InvalidRequestError):
            TestPDU.get_expected_request_data_length(b"\x0d")



"""Holding Registers utilities."""

from typing import Any, Literal, Protocol, TypeVar, cast

from tmodbus.utils.word_aware_struct import WordOrderAwareStruct

from .base import BasePDU
from .holding_registers import RawReadHoldingRegistersPDU, RawReadInputRegistersPDU, RawWriteMultipleRegistersPDU

RT = TypeVar("RT")


class SupportsExecuteAsync(Protocol):
    """Protocol for classes that support the execute method."""

    async def execute(self, pdu: BasePDU[RT]) -> RT:
        """Send the PDU and return the response."""
        ...


class HoldingRegisterReadMixin(SupportsExecuteAsync):
    """Mixin for holding register read operations."""

    word_order: Literal["big", "little"]

    def __init__(self, word_order: Literal["big", "little"] = "big") -> None:
        """Initialize the mixin.

        Args:
            word_order: Word order for multi-register values ('big' or 'little').

        """
        self.word_order = word_order

    async def read_struct_format[RT](
        self,
        start_address: int,
        *,
        format_struct: WordOrderAwareStruct,
        input_register: bool = False,
    ) -> tuple[Any, ...]:
        """Read holding registers and decode them using the provided struct format.

        Args:
            struct_format: Struct format to decode the response
            start_address: Starting address of the registers to read
            format_struct: Struct format to decode the response
            input_register: Whether to read holding registers (False) or input registers (True)

        Returns:
            Decoded response data using the provided struct format

        """
        pdu_class = RawReadInputRegistersPDU if input_register else RawReadHoldingRegistersPDU

        response_bytes = await self.execute(
            pdu_class(
                start_address,
                quantity=format_struct.size // 2,
            )
        )
        return format_struct.unpack_from(response_bytes)

    async def read_simple_struct_format(
        self,
        start_address: int,
        *,
        format_struct: WordOrderAwareStruct,
        input_register: bool = False,
    ) -> Any:
        """Read holding registers and decode them as a single value using the provided struct format.

        Args:
            struct_format: Struct format to decode the response
            start_address: Starting address of the registers to read
            format_struct: Struct format to decode the response
            input_register: Whether to read holding registers (False) or input registers (True)

        Returns:
            Decoded response data as a single value

        """
        return (
            await self.read_struct_format(
                start_address,
                format_struct=format_struct,
                input_register=input_register,
            )
        )[0]

    async def read_uint16(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as an unsigned 16-bit integer.

        An unsigned 16-bit integer is 2 bytes wide (1 register).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded unsigned 16-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">H", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_uint32(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as an unsigned 32-bit integer.

        An unsigned 32-bit integer is 4 bytes wide (2 registers).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded unsigned 32-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">I", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_uint64(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as an unsigned 64-bit integer.

        An unsigned 64-bit integer is 8 bytes wide (4 registers).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded unsigned 64-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">Q", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_int16(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as a signed 16-bit integer.

        A signed 16-bit integer is 2 bytes wide (1 register).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).


        Returns:
            Decoded signed 16-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">h", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_int32(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as a signed 32-bit integer.

        A signed 32-bit integer is 4 bytes wide (2 registers).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded signed 32-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">i", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_int64(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> int:
        """Read holding registers and decode them as a signed 64-bit integer.

        A signed 64-bit integer is 8 bytes wide (4 registers).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded signed 64-bit integer.

        """
        return cast(
            "int",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">q", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_float(
        self,
        start_address: int,
        *,
        input_register: bool = False,
    ) -> float:
        """Read holding registers and decode them as a float.

        A float is 4 bytes wide (2 registers).

        Args:
            start_address: Starting address of the registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).

        Returns:
            Decoded float value.

        """
        return cast(
            "float",
            await self.read_simple_struct_format(
                start_address,
                format_struct=WordOrderAwareStruct(">f", word_order=self.word_order),
                input_register=input_register,
            ),
        )

    async def read_string(
        self,
        start_address: int,
        *,
        number_of_registers: int,
        input_register: bool = False,
        encoding: str = "ascii",
    ) -> str:
        """Read holding registers and decode them as a string.

        Args:
            start_address: Starting address of the registers to read.
            length: Length of the string to decode.
            number_of_registers: Number of registers to read.
            input_register: Whether to read holding registers (False) or input registers (True).
            encoding: Encoding format for the string (default is "ascii").

        Returns:
            Decoded string value.

        """
        format_struct = WordOrderAwareStruct(f">{number_of_registers * 2}s", word_order=self.word_order)
        string_bytes = cast(
            "bytes",
            await self.read_simple_struct_format(
                start_address,
                format_struct=format_struct,
                input_register=input_register,
            ),
        )
        return string_bytes.decode(encoding)


class HoldingRegisterWriteMixin(SupportsExecuteAsync):
    """Mixin for holding register write operations."""

    word_order: Literal["big", "little"]

    def __init__(self, word_order: Literal["big", "little"] = "big") -> None:
        """Initialize the mixin.

        Args:
            word_order: Word order for multi-register values ('big' or 'little').

        """
        self.word_order = word_order

    async def write_struct_format(
        self,
        start_address: int,
        values: tuple[Any, ...],
        *,
        format_struct: WordOrderAwareStruct,
    ) -> int:
        """Write holding registers using the provided struct format.

        Args:
            start_address: Starting address of the registers to write
            values: Values to encode and write
            format_struct: Struct format to encode the values

        Returns:
            The number of registers that have been written.

        """
        return await self.execute(
            RawWriteMultipleRegistersPDU(
                start_address,
                content=format_struct.pack(*values),
            ),
        )

    async def write_simple_struct_format(
        self,
        address: int,
        value: Any,
        *,
        format_struct: WordOrderAwareStruct,
    ) -> Any:
        """Write a single value to holding registers using the provided struct format.

        Args:
            address: Address of the register to write.
            value: Value to encode and write.
            format_struct: Struct format to encode the value.

        Returns:
            None

        """
        return await self.write_struct_format(
            address,
            values=(value,),
            format_struct=format_struct,
        )

    async def write_uint16(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write an unsigned 16-bit integer to holding registers.

        An unsigned 16-bit integer is 1 holding register wide.

        Args:
            address: Address of the register to write.
            value: Unsigned 16-bit integer value to write.

        Returns:
            None

        """
        if not (0 <= value <= 0xFFFF):
            msg = "Value out of range for uint16 (0-65535)."
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">H", word_order=self.word_order),
        )

    async def write_uint32(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write an unsigned 32-bit integer to holding registers.

        An unsigned 32-bit integer is 2 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Unsigned 32-bit integer value to write.

        Returns:
            None

        """
        msg = "Value out of range for uint32 (0-4294967295)."
        if not (0 <= value <= 0xFFFF_FFFF):
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">I", word_order=self.word_order),
        )

    async def write_uint64(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write an unsigned 64-bit integer to holding registers.

        An unsigned 64-bit integer is 4 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Unsigned 64-bit integer value to write.

        Returns:
            None

        """
        if not (0 <= value <= 0xFFFF_FFFF_FFFF_FFFF):
            msg = "Value out of range for uint64 (0-18446744073709551615)."
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">Q", word_order=self.word_order),
        )

    async def write_int16(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write a signed 16-bit integer to holding registers.

        A signed 16-bit integer is 1 holding register wide.

        Args:
            address: Address of the register to write.
            value: Signed 16-bit integer value to write.

        Returns:
            None

        """
        if not (-0x8000 <= value <= 0x7FFF):
            msg = "Value out of range for int16 (-32768 to 32767)."
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">h", word_order=self.word_order),
        )

    async def write_int32(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write a signed 32-bit integer to holding registers.

        A signed 32-bit integer is 2 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Signed 32-bit integer value to write.

        Returns:
            None

        """
        if not (-0x8000_0000 <= value <= 0x7FFF_FFFF):
            msg = "Value out of range for int32 (-2147483648 to 2147483647)."
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">i", word_order=self.word_order),
        )

    async def write_int64(
        self,
        address: int,
        value: int,
    ) -> Any:
        """Write a signed 64-bit integer to holding registers.

        A signed 64-bit integer is 4 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Signed 64-bit integer value to write.

        Returns:
            None

        """
        if not (-0x8000_0000_0000_0000 <= value <= 0x7FFF_FFFF_FFFF_FFFF):
            msg = "Value out of range for int64 (-9223372036854775808 to 9223372036854775807)."
            raise ValueError(msg)

        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">q", word_order=self.word_order),
        )

    async def write_float(
        self,
        address: int,
        value: float,
    ) -> Any:
        """Write a float to holding registers.

        A float is 2 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Float value to write.

        Returns:
            None

        """
        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">f", word_order=self.word_order),
        )

    async def write_double(
        self,
        address: int,
        value: float,
    ) -> Any:
        """Write a double to holding registers.

        A double is 4 holding registers wide.

        Args:
            address: Address of the register to write.
            value: Float value to write.

        Returns:
            None

        """
        return await self.write_simple_struct_format(
            address,
            value,
            format_struct=WordOrderAwareStruct(">d", word_order=self.word_order),
        )

    async def write_string(
        self,
        start_address: int,
        value: str,
        *,
        number_of_registers: int,
        encoding: str = "ascii",
    ) -> Any:
        """Write a string to holding registers.

        Args:
            start_address: Starting address of the registers to write.
            value: String value to write.
            number_of_registers: Number of registers to write.
            encoding: Encoding format for the string (default is "ascii").

        Returns:
            None

        """
        max_length = number_of_registers * 2
        value_bytes = value.encode(encoding)
        if len(value_bytes) > max_length:
            msg = f"String length exceeds maximum size of {max_length} bytes."
            raise ValueError(msg)

        format_struct = WordOrderAwareStruct(f">{number_of_registers * 2}s", word_order=self.word_order)
        # Pad with null bytes if necessary
        value_bytes = value_bytes.ljust(format_struct.size, b"\x00")

        return await self.write_simple_struct_format(
            start_address,
            value_bytes,
            format_struct=format_struct,
        )



"""Example of an asynchronous TCP Modbus client over an SSL connection using tmodbus."""

import asyncio
import ssl

from tmodbus import create_async_tcp_client
from tmodbus.exceptions import InvalidResponseError, ModbusConnectionError, ModbusResponseError


async def example_tcp_client() -> None:
    """Asynchronous TCP Modbus client example."""
    # Replace with your Modbus server's IP and port
    host = "127.0.0.1"
    port = 502

    unit_id = 1  # Modbus unit ID of the target device

    # You can pass any additional parameters supported by `asyncio.open_connection`
    # For example, to use an SSL context, you can set `ssl=True` or provide a custom SSLContext

    # The create_async_tcp_client function returns an instance of AsyncModbusClient
    client = create_async_tcp_client(host, port, unit_id=unit_id, ssl=True)

    try:
        await client.connect()
        # Read 2 holding registers starting at address 100
        coils_response = await client.read_holding_registers(start_address=100, quantity=2)

        print("Contents of holding registers 100 and 101: ", coils_response)

        # Write value 123 to holding register at address 1
        await client.write_single_register(address=1, value=123)

        # Write values [10, 20, 30] to holding registers starting at address 10
        await client.write_multiple_registers(start_address=10, values=[10, 20, 30])

    except ModbusResponseError as e:
        print(f"The server responded with error code {e.error_code:#04x} for function {e.function_code:#04x}")
    except InvalidResponseError as e:
        print(f"Received invalid response: {e}")
    except ModbusConnectionError as e:
        print(f"A connection error occurred: {e}")
    finally:
        await client.disconnect()

    # Alternatively, you can use the client as an async context manager
    # which automatically handles connection and disconnection.
    # We also demonstrate passing of a custom SSL context here.

    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ssl_context.load_default_certs(purpose=ssl.Purpose.SERVER_AUTH)
    ssl_context.check_hostname = True
    ssl_context.verify_mode = ssl.CERT_REQUIRED

    async with create_async_tcp_client(host, port, unit_id=unit_id, ssl=ssl_context) as client2:
        print("Status of coils 0-7: ", await client2.read_coils(start_address=0, quantity=8))


if __name__ == "__main__":
    asyncio.run(example_tcp_client())



"""Transport layer."""

from .async_base import AsyncBaseTransport
from .async_rtu import AsyncRtuTransport
from .async_smart import AsyncSmartTransport
from .async_tcp import AsyncTcpTransport

__all__ = [
    "AsyncBaseTransport",
    "AsyncRtuTransport",
    "AsyncSmartTransport",
    "AsyncTcpTransport",
]



"""Tests for holding_registers_struct module."""

import math
import struct
from typing import Literal
from unittest.mock import AsyncMock

import pytest
from tmodbus.pdu.base import BaseClientPDU
from tmodbus.pdu.holding_registers_struct import HoldingRegisterReadMixin, HoldingRegisterWriteMixin
from tmodbus.utils.word_aware_struct import WordOrderAwareStruct


class MockClient(HoldingRegisterReadMixin, HoldingRegisterWriteMixin):
    """Mock client for testing mixins."""

    execute: AsyncMock

    def __init__(self, word_order: Literal["big", "little"] = "big") -> None:
        """Mock Client."""
        HoldingRegisterReadMixin.__init__(self, word_order=word_order)
        HoldingRegisterWriteMixin.__init__(self, word_order=word_order)
        self.execute = AsyncMock()


class TestHoldingRegisterReadMixin:
    """Test HoldingRegisterReadMixin class."""

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_struct_format(self, word_order: Literal["big", "little"]) -> None:
        """Test read_struct_format with both word orders."""
        client = MockClient(word_order=word_order)
        # Mock response: 4 bytes representing uint32
        client.execute.return_value = b"\x0a\x0b\x0c\x0d"

        format_struct = WordOrderAwareStruct(">I", word_order=word_order)
        result = await client.read_struct_format(100, format_struct=format_struct)

        # Verify execute was called with correct PDU
        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.start_address == 100
        assert pdu.quantity == 2  # 4 bytes = 2 registers

        # Verify result
        assert isinstance(result, tuple)
        assert len(result) == 1

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_struct_format_input_register(self, word_order: Literal["big", "little"]) -> None:
        """Test read_struct_format with input registers."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\x01\x02"

        format_struct = WordOrderAwareStruct(">H", word_order=word_order)
        result = await client.read_struct_format(50, format_struct=format_struct, input_register=True)

        assert client.execute.called
        assert result == (0x0102,)

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_simple_struct_format(self, word_order: Literal["big", "little"]) -> None:
        """Test read_simple_struct_format returns single value."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\x12\x34"

        format_struct = WordOrderAwareStruct(">H", word_order=word_order)
        result = await client.read_simple_struct_format(100, format_struct=format_struct)

        # Should return single value, not tuple
        assert result == 0x1234

    @pytest.mark.parametrize(
        ("word_order", "expected"),
        [
            ("big", 0x1234),
            ("little", 0x1234),  # Single register, no swapping
        ],
    )
    async def test_read_uint16(self, word_order: Literal["big", "little"], expected: int) -> None:
        """Test read_uint16 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\x12\x34"

        result = await client.read_uint16(100)

        assert result == expected
        assert client.execute.called

    @pytest.mark.parametrize(
        ("word_order", "response", "expected"),
        [
            ("big", b"\x0a\x0b\x0c\x0d", 0x0A0B0C0D),
            ("little", b"\x0c\x0d\x0a\x0b", 0x0A0B0C0D),
        ],
    )
    async def test_read_uint32(self, word_order: Literal["big", "little"], response: bytes, expected: int) -> None:
        """Test read_uint32 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = response

        result = await client.read_uint32(100)

        assert result == expected

    @pytest.mark.parametrize(
        ("word_order", "response", "expected"),
        [
            ("big", b"\x0a\x0b\x0c\x0d\x0e\x0f\x01\x02", 0x0A0B0C0D0E0F0102),
            ("little", b"\x01\x02\x0e\x0f\x0c\x0d\x0a\x0b", 0x0A0B0C0D0E0F0102),
        ],
    )
    async def test_read_uint64(self, word_order: Literal["big", "little"], response: bytes, expected: int) -> None:
        """Test read_uint64 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = response

        result = await client.read_uint64(100)

        assert result == expected

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_int16(self, word_order: Literal["big", "little"]) -> None:
        """Test read_int16 with negative value."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\xff\xff"  # -1

        result = await client.read_int16(100)

        assert result == -1

    @pytest.mark.parametrize(
        ("word_order", "response", "expected"),
        [
            ("big", b"\x0a\x0b\x0c\x0d", 0x0A0B0C0D),
            ("little", b"\x0c\x0d\x0a\x0b", 0x0A0B0C0D),
            ("big", b"\xff\xff\xff\xff", -1),
            ("little", b"\xff\xff\xff\xff", -1),
            ("big", b"\xff\xff\xff\xfe", -2),
            ("little", b"\xff\xfe\xff\xff", -2),
            ("big", b"\x80\x00\x00\x00", -2147483648),  # Min int32
            ("little", b"\x00\x00\x80\x00", -2147483648),
            ("big", b"\x7f\xff\xff\xff", 2147483647),  # Max int32
            ("little", b"\xff\xff\x7f\xff", 2147483647),
        ],
    )
    async def test_read_int32(self, word_order: Literal["big", "little"], response: bytes, expected: int) -> None:
        """Test read_int32 with both word orders and negative values."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = response

        result = await client.read_int32(100)

        assert result == expected

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_int64(self, word_order: Literal["big", "little"]) -> None:
        """Test read_int64 with negative value."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\xff\xff\xff\xff\xff\xff\xff\xff"  # -1

        result = await client.read_int64(100)

        assert result == -1

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_float(self, word_order: Literal["big", "little"]) -> None:
        """Test read_float with both word orders."""
        client = MockClient(word_order=word_order)
        # Encode 1.0 as float and apply word order

        float_bytes = struct.pack(">f", 1.0)
        if word_order == "little":
            # Swap register order for little endian
            float_bytes = float_bytes[2:4] + float_bytes[0:2]

        client.execute.return_value = float_bytes

        result = await client.read_float(100)

        assert abs(result - 1.0) < 0.0001  # Float comparison with tolerance

    @pytest.mark.parametrize(
        ("word_order", "response"),
        [
            ("big", b"HELLO\x00\x00\x00"),
            ("little", b"\x00\x00O\x00LLHE"),  # Word-swapped format from device
        ],
    )
    async def test_read_string(self, word_order: Literal["big", "little"], response: bytes) -> None:
        """Test read_string with both word orders."""
        client = MockClient(word_order=word_order)
        # 4 registers = 8 bytes
        client.execute.return_value = response

        result = await client.read_string(100, number_of_registers=4)

        assert "HELLO" in result

    @pytest.mark.parametrize(
        ("word_order", "response"),
        [
            ("big", b"TEST\x00\x00\x00\x00"),
            ("little", b"\x00\x00\x00\x00STTE"),  # Word-swapped format from device
        ],
    )
    async def test_read_string_with_encoding(self, word_order: Literal["big", "little"], response: bytes) -> None:
        """Test read_string with custom encoding."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = response

        result = await client.read_string(100, number_of_registers=4, encoding="utf-8")

        assert "TEST" in result

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_read_input_register_flag(self, word_order: Literal["big", "little"]) -> None:
        """Test that input_register flag is passed through correctly."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = b"\x12\x34"

        await client.read_uint16(100, input_register=True)

        # Verify the PDU type by checking the execute call
        pdu = client.execute.call_args[0][0]
        # Input register PDU should be different from holding register PDU
        assert pdu.__class__.__name__ == "RawReadInputRegistersPDU"


class TestHoldingRegisterWriteMixin:
    """Test HoldingRegisterWriteMixin class."""

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_write_struct_format(self, word_order: Literal["big", "little"]) -> None:
        """Test write_struct_format with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 2  # Number of registers written

        format_struct = WordOrderAwareStruct(">I", word_order=word_order)
        result = await client.write_struct_format(100, (0x0A0B0C0D,), format_struct=format_struct)

        assert result == 2
        assert client.execute.called

        # Verify PDU
        pdu = client.execute.call_args[0][0]
        assert pdu.start_address == 100
        assert len(pdu.content) == 4  # 4 bytes for uint32

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_write_simple_struct_format(self, word_order: Literal["big", "little"]) -> None:
        """Test write_simple_struct_format."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 1

        format_struct = WordOrderAwareStruct(">H", word_order=word_order)
        result = await client.write_simple_struct_format(100, 0x1234, format_struct=format_struct)

        assert result == 1
        assert client.execute.called

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\x12\x34"),
            ("little", b"\x12\x34"),  # Single register, no swapping
        ],
    )
    async def test_write_uint16(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_uint16 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 1

        await client.write_uint16(100, 0x1234)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_uint16_out_of_range_low(self) -> None:
        """Test write_uint16 with value below range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for uint16"):
            await client.write_uint16(100, -1)

    async def test_write_uint16_out_of_range_high(self) -> None:
        """Test write_uint16 with value above range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for uint16"):
            await client.write_uint16(100, 0x10000)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\x12\x34\x56\x78"),
            ("little", b"\x56\x78\x12\x34"),  # Swapped register order
        ],
    )
    async def test_write_uint32(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_uint32 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 2

        await client.write_uint32(100, 0x12345678)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_uint32_out_of_range(self) -> None:
        """Test write_uint32 with value out of range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for uint32"):
            await client.write_uint32(100, 0x1_0000_0000)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\x12\x34\x56\x78\x9a\xbc\xde\xf0"),
            ("little", b"\xde\xf0\x9a\xbc\x56\x78\x12\x34"),  # Swapped register order
        ],
    )
    async def test_write_uint64(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_uint64 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 4

        await client.write_uint64(100, 0x123456789ABCDEF0)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_uint64_out_of_range(self) -> None:
        """Test write_uint64 with value out of range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for uint64"):
            await client.write_uint64(100, 0x1_0000_0000_0000_0000)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\xfb\x2e"),  # -1234 in signed 16-bit
            ("little", b"\xfb\x2e"),  # Single register, no swapping
        ],
    )
    async def test_write_int16(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_int16 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 1

        await client.write_int16(100, -1234)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_int16_out_of_range_low(self) -> None:
        """Test write_int16 with value below range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for int16"):
            await client.write_int16(100, -0x8001)

    async def test_write_int16_out_of_range_high(self) -> None:
        """Test write_int16 with value above range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for int16"):
            await client.write_int16(100, 0x8000)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\xff\xfe\x1d\xc0"),  # -123456 in signed 32-bit
            ("little", b"\x1d\xc0\xff\xfe"),  # Swapped register order
        ],
    )
    async def test_write_int32(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_int32 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 2

        await client.write_int32(100, -123456)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_int32_out_of_range(self) -> None:
        """Test write_int32 with value out of range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for int32"):
            await client.write_int32(100, -0x8000_0001)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\xff\xff\xff\xff\xf8\xa4\x32\xeb"),  # -123456789 in signed 64-bit
            ("little", b"\x32\xeb\xf8\xa4\xff\xff\xff\xff"),  # Swapped register order
        ],
    )
    async def test_write_int64(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_int64 with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 4

        await client.write_int64(100, -123456789)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_int64_out_of_range(self) -> None:
        """Test write_int64 with value out of range."""
        client = MockClient()

        with pytest.raises(ValueError, match="Value out of range for int64"):
            await client.write_int64(100, -0x8000_0000_0000_0001)

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\x40\x49\x0f\xd0"),  # 3.14159 as float
            ("little", b"\x0f\xd0\x40\x49"),  # Swapped register order
        ],
    )
    async def test_write_float(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_float with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 2

        await client.write_float(100, 3.14159)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"\x40\x09\x21\xfb\x54\x44\x2d\x18"),  # pi as double
            ("little", b"\x2d\x18\x54\x44\x21\xfb\x40\x09"),  # Swapped register order
        ],
    )
    async def test_write_double(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_double with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 4

        await client.write_double(100, 3.141592653589793)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    @pytest.mark.parametrize(
        ("word_order", "expected_bytes"),
        [
            ("big", b"HELLO\x00\x00\x00"),  # Right-padded with nulls (ljust)
            ("little", b"\x00\x00O\x00LLHE"),  # Swapped register order
        ],
    )
    async def test_write_string(self, word_order: Literal["big", "little"], expected_bytes: bytes) -> None:
        """Test write_string with both word orders."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 4

        await client.write_string(100, "HELLO", number_of_registers=4)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        assert pdu.content == expected_bytes

    async def test_write_string_too_long(self) -> None:
        """Test write_string with string too long for register count."""
        client = MockClient()

        with pytest.raises(ValueError, match="String length exceeds maximum size"):
            await client.write_string(100, "VERYLONGSTRING", number_of_registers=2)

    @pytest.mark.parametrize("word_order", ["big", "little"])
    async def test_write_string_with_encoding(self, word_order: Literal["big", "little"]) -> None:
        """Test write_string with custom encoding."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 3

        await client.write_string(100, "TEST", number_of_registers=3, encoding="utf-8")

        assert client.execute.called

    @pytest.mark.parametrize(
        ("word_order", "bytes_result"),
        [
            ("big", b"TEST\x00\x00\x00\x00"),
            ("little", b"\x00\x00\x00\x00STTE"),
        ],
    )
    async def test_write_string_padding(self, word_order: Literal["big", "little"], bytes_result: bytes) -> None:
        """Test that short strings are padded correctly."""
        client = MockClient(word_order=word_order)
        client.execute.return_value = 4

        await client.write_string(100, "TEST", number_of_registers=4)

        assert client.execute.called
        pdu = client.execute.call_args[0][0]
        # Should be padded to 8 bytes (4 registers)
        assert len(pdu.content) == 8
        # Verify content contains the string (word order may affect byte arrangement)
        assert pdu.content == bytes_result


class TestRoundTripReadWrite:
    """Test round-trip write then read operations."""

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            0,
            1,
            100,
            1234,
            32767,  # Max positive int16
            65535,  # Max uint16
            42,  # Random value
            0x1234,  # Specific hex pattern
        ],
    )
    async def test_roundtrip_uint16(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back uint16 values."""
        client = MockClient(word_order=word_order)

        # Capture the written bytes
        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 1

        client.execute.side_effect = capture_write
        await client.write_uint16(100, value)

        # Now read it back
        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_uint16(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            -32768,  # Min int16
            -1,
            0,
            1,
            100,
            32767,  # Max int16
            -12345,  # Random negative
            12345,  # Random positive
        ],
    )
    async def test_roundtrip_int16(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back int16 values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 1

        client.execute.side_effect = capture_write
        await client.write_int16(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_int16(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            0,
            1,
            1000,
            0x12345678,
            0xFFFFFFFF,  # Max uint32
            0x80000000,  # 2^31
            0xDEADBEEF,  # Common test pattern
        ],
    )
    async def test_roundtrip_uint32(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back uint32 values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 2

        client.execute.side_effect = capture_write
        await client.write_uint32(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_uint32(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            -2147483648,  # Min int32
            -1,
            0,
            1,
            2147483647,  # Max int32
            -123456789,  # Random negative
            123456789,  # Random positive
            -2147483647,  # Min + 1
        ],
    )
    async def test_roundtrip_int32(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back int32 values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 2

        client.execute.side_effect = capture_write
        await client.write_int32(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_int32(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            0,
            1,
            0xFFFFFFFFFFFFFFFF,  # Max uint64
            0x123456789ABCDEF0,
            0x8000000000000000,  # 2^63
            9223372036854775807,  # Max int64 as uint
            1234567890123456789,  # Random large value
        ],
    )
    async def test_roundtrip_uint64(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back uint64 values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 4

        client.execute.side_effect = capture_write
        await client.write_uint64(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_uint64(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            -9223372036854775808,  # Min int64
            -1,
            0,
            1,
            9223372036854775807,  # Max int64
            -1234567890123456789,  # Random negative
            1234567890123456789,  # Random positive
        ],
    )
    async def test_roundtrip_int64(self, word_order: Literal["big", "little"], value: int) -> None:
        """Test writing and reading back int64 values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[int]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 4

        client.execute.side_effect = capture_write
        await client.write_int64(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_int64(100)

        assert result == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        "value",
        [
            0.0,
            1.0,
            -1.0,
            3.14159,
            -3.14159,
            123.456,
            -999.999,
            1.23e-10,  # Very small positive
            -1.23e-10,  # Very small negative
            1.23e10,  # Large positive
            -1.23e10,  # Large negative
            float("inf"),  # Infinity
            float("-inf"),  # Negative infinity
        ],
    )
    async def test_roundtrip_float(self, word_order: Literal["big", "little"], value: float) -> None:
        """Test writing and reading back float values."""
        client = MockClient(word_order=word_order)

        written_bytes = None

        def capture_write(pdu: BaseClientPDU[float]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return 2

        client.execute.side_effect = capture_write
        await client.write_float(100, value)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_float(100)

        # Handle special float values
        if math.isnan(value):  # NaN check
            assert math.isnan(result)
        elif value == float("inf"):
            assert result == float("inf")
        elif value == float("-inf"):
            assert result == float("-inf")
        else:
            # Float comparison with tolerance
            assert abs(result - value) < abs(value * 1e-6) if value != 0 else abs(result) < 1e-6

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        ("value", "num_registers"),
        [
            ("", 2),  # Empty string
            ("A", 2),  # Single char
            ("TEST", 2),  # Exact fit (4 bytes = 2 registers)
            ("HELLO", 4),  # Padded
            ("ABCDEFGH", 4),  # Exact fit (8 bytes = 4 registers)
            ("Short", 5),  # Padded (10 bytes = 5 registers)
            ("X" * 20, 10),  # Long string
            ("123", 3),  # Numeric string
            ("Hello World!", 7),  # With space and punctuation
        ],
    )
    async def test_roundtrip_string(self, word_order: Literal["big", "little"], value: str, num_registers: int) -> None:
        """Test writing and reading back string values."""
        # Skip if string is too long for the number of registers
        if len(value.encode("utf-8")) > num_registers * 2:
            pytest.skip(f"String '{value}' too long for {num_registers} registers")

        client = MockClient(word_order=word_order)

        written_bytes = None

        def _capture_write(pdu: BaseClientPDU[str]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return num_registers

        client.execute.side_effect = _capture_write
        await client.write_string(100, value, number_of_registers=num_registers)

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_string(100, number_of_registers=num_registers)

        # Strip null bytes and whitespace for comparison
        assert result.rstrip("\x00") == value

    @pytest.mark.parametrize("word_order", ["big", "little"])
    @pytest.mark.parametrize(
        ("value", "num_registers"),
        [
            ("Café", 4),  # UTF-8 special character
            ("Hello 世界", 10),  # UTF-8 Chinese characters
        ],
    )
    async def test_roundtrip_string_utf8(
        self, word_order: Literal["big", "little"], value: str, num_registers: int
    ) -> None:
        """Test writing and reading back UTF-8 string values."""
        # Skip if string is too long for the number of registers
        if len(value.encode("utf-8")) > num_registers * 2:
            pytest.skip(f"String '{value}' too long for {num_registers} registers")

        client = MockClient(word_order=word_order)

        written_bytes = None

        def _capture_write(pdu: BaseClientPDU[str]) -> int:
            nonlocal written_bytes
            written_bytes = pdu.content  # type: ignore[attr-defined]
            return num_registers

        client.execute.side_effect = _capture_write
        await client.write_string(100, value, number_of_registers=num_registers, encoding="utf-8")

        client.execute.side_effect = None
        client.execute.return_value = written_bytes
        result = await client.read_string(100, number_of_registers=num_registers, encoding="utf-8")

        # Strip null bytes and whitespace for comparison
        assert result.rstrip("\x00") == value



"""Read Device Identification PDU.

cfr. section 6.21 of the Modbus Application Protocol Specification V1.1b3
"""

import logging
import struct
from dataclasses import dataclass
from enum import IntEnum
from typing import Literal

from tmodbus.const import FunctionCode

from .base import BaseSubFunctionClientPDU

logger = logging.getLogger(__name__)


class ObjectName(IntEnum):
    """Object ID to Object Name mapping."""

    VENDOR_NAME = 0x00  # Basic, Mandatory
    PRODUCT_CODE = 0x01  # Basic, Mandatory
    MAJOR_MINOR_REVISION = 0x02  # Basic, Mandatory
    VENDOR_URL = 0x03  # Regular, Optional
    PRODUCT_NAME = 0x04  # Regular, Optional
    MODEL_NAME = 0x05  # Regular, Optional
    USER_APPLICATION_NAME = 0x06  # Regular, Optional

    # 0x07 to 0x7F: Reserved for regular, optional objects
    # 0x80 to 0xFF: Reserved for extended (manufacturer-specific), optional objects


class ConformityLevel(IntEnum):
    """Conformity Level."""

    BASIC = 0x01
    """Basic identification (stream access only)"""
    REGULAR = 0x02
    """Regular identification (stream access only)"""
    EXTENDED = 0x03
    """Extended identification (stream access only)"""
    BASIC_PLUS = 0x81
    """Basic identification (stream access and individual access)"""
    REGULAR_PLUS = 0x82
    """Regular identification (stream access and individual access)"""
    EXTENDED_PLUS = 0x83
    """Extended identification (stream access and individual access)"""


@dataclass(frozen=True)
class ReadDeviceIdentificationResponse:
    """Contents of the ReadDeviceInfo response."""

    device_id_code: Literal[0x01, 0x02, 0x03, 0x04]
    conformity_level: ConformityLevel
    more: bool
    next_object_id: int
    number_of_objects: int

    objects: dict[int, bytes]


@dataclass(frozen=True)
class ReadDeviceIdentificationPDU(BaseSubFunctionClientPDU[ReadDeviceIdentificationResponse]):
    """Modbus Request to read a device identifier."""

    function_code = FunctionCode.ENCAPSULATED_INTERFACE_TRANSPORT

    sub_function_code = 0x0E
    read_device_id_code: Literal[0x01, 0x02, 0x03, 0x04]
    object_id: int

    def __post_init__(self) -> None:
        """Validate ReadDeviceIdentificationPDU."""
        if not (0x00 <= self.object_id < 0xFF):
            msg = "Object ID must be between 0x00 and 0xFF."
            raise ValueError(msg)

    def encode_request(self) -> bytes:
        """Encode ReadDeviceIdentifierPDU."""
        return struct.pack(
            ">BBBB",
            self.function_code,
            self.sub_function_code,
            self.read_device_id_code,
            self.object_id,
        )

    def decode_response(self, response: bytes) -> ReadDeviceIdentificationResponse:
        """Decode Device Identifier PDU response."""
        response_header_struct = struct.Struct(">BBBBBBB")
        (
            function_code,
            sub_function_code,
            device_id_code,
            conformity_level,
            more,
            next_object_id,
            number_of_objects,
        ) = response_header_struct.unpack_from(response, 0)

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:#04x}, received {function_code:#04x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = (
                f"Invalid sub function code: expected {self.sub_function_code:#04x}, received {sub_function_code:#04x}"
            )
            raise ValueError(msg)

        if more not in (0x00, 0xFF):
            msg = f"Invalid 'more' value: {more:#04x}"
            raise ValueError(msg)

        objects: dict[int, bytes] = {}
        offset = response_header_struct.size
        while offset < len(response):
            obj_id, obj_length = struct.unpack_from(">BB", response, offset)
            offset += 2
            objects[obj_id] = response[offset : offset + obj_length]
            offset += obj_length

        if offset != len(response):
            logger.warning("Response has %d extra bytes", len(response) - offset)

        return ReadDeviceIdentificationResponse(
            device_id_code=device_id_code,
            conformity_level=ConformityLevel(conformity_level),
            more=bool(more),
            next_object_id=next_object_id,
            number_of_objects=number_of_objects,
            objects=objects,
        )



"""Tests for tmodbus/pdu/pdu.py ."""

import pytest
from tmodbus.const import FunctionCode
from tmodbus.pdu import (
    BaseClientPDU,
    BaseSubFunctionClientPDU,
    get_pdu_class,
    register_pdu_class,
)


class TestGetPDUClass:
    """Tests for get_pdu_class function."""

    def test_get_pdu_class_valid_function_code(self) -> None:
        """Test getting PDU class by valid function code."""
        pdu_class = get_pdu_class(bytes([int(FunctionCode.READ_HOLDING_REGISTERS)]))
        assert pdu_class.__name__ == "ReadHoldingRegistersPDU"

    def test_get_pdu_class_valid_function_code_as_int(self) -> None:
        """Test getting PDU class by function code as int."""
        pdu_class = get_pdu_class(bytes([0x03]))  # FunctionCode.READ_HOLDING_REGISTERS
        assert pdu_class.__name__ == "ReadHoldingRegistersPDU"

    def test_get_pdu_class_unknown_function_code(self) -> None:
        """Test unknown function code raises ValueError."""
        with pytest.raises(ValueError, match="Unsupported function code: 0x99"):
            get_pdu_class(bytes([0x99]))

    def test_get_pdu_class_unsupported_function_code(self) -> None:
        """Test unsupported function code raises ValueError."""
        with pytest.raises(ValueError, match="Unsupported function code: 0x18"):
            get_pdu_class(bytes([FunctionCode.READ_FIFO_QUEUE]))

    def test_get_pdu_class_with_sub_function(self) -> None:
        """Test getting PDU class with sub-function code."""
        # ReadDeviceIdentificationPDU uses function code 0x2B and sub-function code 0x0E
        pdu_class = get_pdu_class(bytes([0x2B, 0x0E]))
        assert pdu_class.__name__ == "ReadDeviceIdentificationPDU"

    def test_get_pdu_class_unknown_sub_function_code(self) -> None:
        """Test unknown sub-function code raises ValueError."""
        # Function code 0x2B is valid, but sub-function 0xFF is not registered
        with pytest.raises(
            ValueError,
            match="Unsupported sub-function code: 0xff for function code 0x2b",
        ):
            get_pdu_class(bytes([0x2B, 0xFF]))


class TestRegisterPDUClass:
    """Tests for register_pdu_class function."""

    def test_register_normal_pdu_class(self) -> None:
        """Test registering a normal PDU class."""

        class CustomPDU(BaseClientPDU[int]):
            function_code = 0xF0

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register the custom PDU
        register_pdu_class(CustomPDU)

        # Verify it can be retrieved
        pdu_class = get_pdu_class(bytes([0xF0]))
        assert pdu_class == CustomPDU

    def test_register_duplicate_normal_pdu_class(self) -> None:
        """Test registering a duplicate function code raises ValueError."""

        class CustomPDU1(BaseClientPDU[int]):
            function_code = 0xF1

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class CustomPDU2(BaseClientPDU[int]):
            function_code = 0xF1

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register first PDU
        register_pdu_class(CustomPDU1)

        # Try to register duplicate
        with pytest.raises(
            ValueError,
            match=r"Function code 0xf1 is already registered to CustomPDU1",
        ):
            register_pdu_class(CustomPDU2)

    def test_register_sub_function_pdu_class(self) -> None:
        """Test registering a sub-function PDU class."""

        class CustomSubFunctionPDU(BaseSubFunctionClientPDU[int]):
            function_code = 0xF2
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register the custom sub-function PDU
        register_pdu_class(CustomSubFunctionPDU)

        # Verify it can be retrieved
        pdu_class = get_pdu_class(bytes([0xF2, 0x01]))
        assert pdu_class == CustomSubFunctionPDU

    def test_register_duplicate_sub_function_pdu_class(self) -> None:
        """Test registering a duplicate sub-function code raises ValueError."""

        class CustomSubFunctionPDU1(BaseSubFunctionClientPDU[int]):
            function_code = 0xF3
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class CustomSubFunctionPDU2(BaseSubFunctionClientPDU[int]):
            function_code = 0xF3
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register first sub-function PDU
        register_pdu_class(CustomSubFunctionPDU1)

        # Try to register duplicate
        with pytest.raises(
            ValueError,
            match=(
                r"A PDU with function code 0xf3, "
                r"and sub-function code 0x01 is already registered: "
                r"CustomSubFunctionPDU1\."
            ),
        ):
            register_pdu_class(CustomSubFunctionPDU2)

    def test_register_sub_function_pdu_when_normal_exists(self) -> None:
        """Test registering a sub-function PDU when a normal PDU already exists."""

        class NormalPDU(BaseClientPDU[int]):
            function_code = 0xF4

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class SubFunctionPDU(BaseSubFunctionClientPDU[int]):
            function_code = 0xF4
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register normal PDU first
        register_pdu_class(NormalPDU)

        # Try to register sub-function PDU with same function code
        with pytest.raises(
            ValueError,
            match=(
                r"Function code 0xf4 is already registered "
                r"for a non-subfunction PDU NormalPDU\."
            ),
        ):
            register_pdu_class(SubFunctionPDU)

    def test_register_normal_pdu_when_sub_function_exists(self) -> None:
        """Test registering a normal PDU when sub-function PDUs already exist."""

        class SubFunctionPDU1(BaseSubFunctionClientPDU[int]):
            function_code = 0xF5
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class SubFunctionPDU2(BaseSubFunctionClientPDU[int]):
            function_code = 0xF5
            sub_function_code = 0x02

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class NormalPDU(BaseClientPDU[int]):
            function_code = 0xF5

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register sub-function PDUs first
        register_pdu_class(SubFunctionPDU1)
        register_pdu_class(SubFunctionPDU2)

        # Try to register normal PDU with same function code
        with pytest.raises(
            ValueError,
            match=(
                r"Function code 0xf5 is already registered with sub-functions: "
                r"0x01: SubFunctionPDU1, 0x02: SubFunctionPDU2"
            ),
        ):
            register_pdu_class(NormalPDU)

    def test_register_multiple_sub_functions_same_function_code(self) -> None:
        """Test registering multiple sub-function PDUs with the same function code."""

        class SubFunctionPDUA(BaseSubFunctionClientPDU[int]):
            function_code = 0xF6
            sub_function_code = 0x01

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        class SubFunctionPDUB(BaseSubFunctionClientPDU[int]):
            function_code = 0xF6
            sub_function_code = 0x02

            def encode_request(self) -> bytes:
                return b""

            def decode_response(self, _response: bytes) -> int:
                return 0

        # Register both sub-function PDUs
        register_pdu_class(SubFunctionPDUA)
        register_pdu_class(SubFunctionPDUB)

        # Verify both can be retrieved
        pdu_class_a = get_pdu_class(bytes([0xF6, 0x01]))
        pdu_class_b = get_pdu_class(bytes([0xF6, 0x02]))
        assert pdu_class_a == SubFunctionPDUA
        assert pdu_class_b == SubFunctionPDUB



r"""Word order aware struct for Modbus multi-register values.

In Modbus, the byte order within a register (which is 2 bytes long) is defined as big-endian.

However, the Modbus standard did not define how values spanning multiple registers should be ordered.
This is called "word order" and can be either "big" (most significant register first) or "little"
(least significant register first).

This module provides a struct-like class that is aware of word order when packing and unpacking.

It extends the standard library's struct.Struct class and adds a word_order parameter to the constructor.

Example:
    >>> from tmodbus.utils.word_aware_struct import WordOrderAwareStruct
    >>> s = WordOrderAwareStruct(">I", word_order="little")  # 32-bit unsigned int, little-endian word order
    >>> data = s.pack(0x0A0B0C0D)
    >>> data
    b'\x0c\x0d\x0a\x0b' # the registers are swapped, the bytes within each register are not
    >>> result = s.unpack(data)
    >>> hex(result[0])
    '0x0a0b0c0d'

"""

import re
import struct
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer, WriteableBuffer

BYTES_PER_REGISTER = 2


class WordOrderAwareStruct(struct.Struct):
    """Struct with word order awareness for multi-register values."""

    _value_lengths: list[int] | None = None

    def __init__(
        self,
        format: str,  # noqa: A002
        word_order: Literal["big", "little"] = "big",
    ) -> None:
        """Initialize the struct.

        Args:
            format: Struct format string.
            word_order: Word order for multi-register values ('big' or 'little').

        """
        super().__init__(format)

        if self.size % 2 != 0:
            msg = "Struct size must be a multiple of 2 bytes: Modbus holding registers are two bytes a piece."
            raise ValueError(msg)

        self.word_order = word_order
        if word_order == "little":
            self._value_lengths = WordOrderAwareStruct.parse_format_lengths(format)

    @staticmethod
    def parse_format_lengths(format_str: str) -> list[int]:
        """Parse a struct format string and return a list of value lengths in bytes."""
        # Remove byte order character if present
        if format_str and format_str[0] in "@=<>!":
            format_str = format_str[1:]

        # Regex to match format tokens (e.g., '2H', 'f', '10s')
        token_re = re.compile(r"(\d*)([xcbB?hHiIlLqQnNefdspP])")
        lengths = []
        for count_str, code in token_re.findall(format_str):
            count = int(count_str) if count_str else 1
            if code in ("s", "p"):
                # 's' and 'p' are special cases: their count indicates the total size in bytes
                lengths.append(count)
            else:
                size = struct.calcsize(code)
                lengths.extend([size] * count)

        return lengths

    def _swap_word_order(self, data: "ReadableBuffer") -> bytes:
        """Swap the word order of the data if needed."""
        if not self._value_lengths:
            # big endian word order, no need to swap
            return bytes(data)

        # Swap the order of 2-byte registers, but do not reverse bytes within each register
        start_idx = 0
        swapped_data = bytearray(data)

        current_length = 0

        for length in self._value_lengths:
            current_length += length
            if current_length % 2 != 0:
                # we need an even number of bytes to swap registers
                continue

            if current_length == BYTES_PER_REGISTER:
                # nothing to swap for single-register values
                start_idx += length
                current_length = 0
                continue

            registers_to_swap = length // BYTES_PER_REGISTER

            for swap_idx in range(registers_to_swap):
                src_idx = start_idx + swap_idx * BYTES_PER_REGISTER
                dst_idx = start_idx + current_length - swap_idx * BYTES_PER_REGISTER - BYTES_PER_REGISTER

                swapped_data[dst_idx : dst_idx + BYTES_PER_REGISTER] = data[src_idx : src_idx + BYTES_PER_REGISTER]  # type: ignore[index]

            start_idx += current_length
            current_length = 0

        return bytes(swapped_data)

    def unpack(self, buffer: "ReadableBuffer") -> tuple[Any, ...]:
        """Unpack buffer with word order consideration."""
        return super().unpack(self._swap_word_order(buffer))

    def unpack_from(self, buffer: "ReadableBuffer", offset: int = 0) -> tuple[Any, ...]:
        """Unpack from buffer with word order consideration."""
        return super().unpack_from(self._swap_word_order(buffer), offset)

    def pack(self, *args: Any) -> bytes:
        """Pack values into bytes with word order consideration."""
        return self._swap_word_order(super().pack(*args))

    def pack_into(self, buffer: "WriteableBuffer", offset: int, *args: Any) -> None:
        """Pack values into buffer with word order consideration."""
        packed_data = self._swap_word_order(super().pack(*args))
        buffer[offset : offset + len(packed_data)] = packed_data  # type: ignore[index]



"""Tests for word_aware_struct module."""

import struct
from typing import Any, Literal

import pytest
from tmodbus.utils.word_aware_struct import WordOrderAwareStruct


def test_init_odd_size_raises() -> None:
    """Test that odd-sized structs raise ValueError."""
    with pytest.raises(ValueError, match="multiple of 2 bytes"):
        WordOrderAwareStruct(">b")  # 1 byte


@pytest.mark.parametrize(
    ("format_str", "word_order", "value", "expected_bytes"),
    [
        # Single register (2 bytes) - no swapping needed
        (">H", "big", 0x0102, b"\x01\x02"),
        (">H", "little", 0x0102, b"\x01\x02"),
        # 32-bit values (2 registers)
        (">I", "big", 0x0A0B0C0D, b"\x0a\x0b\x0c\x0d"),
        (">I", "little", 0x0A0B0C0D, b"\x0c\x0d\x0a\x0b"),
        (">f", "big", 1.0, b"\x3f\x80\x00\x00"),
        (">f", "little", 1.0, b"\x00\x00\x3f\x80"),  # swapped register order
        # 64-bit values (4 registers)
        (">Q", "big", 0x0A0B0C0D0E0F0102, b"\x0a\x0b\x0c\x0d\x0e\x0f\x01\x02"),
        (">Q", "little", 0x0A0B0C0D0E0F0102, b"\x01\x02\x0e\x0f\x0c\x0d\x0a\x0b"),
        (">d", "big", 1.0, struct.pack(">d", 1.0)),
    ],
)
def test_pack_single_value(
    format_str: str,
    word_order: Literal["little", "big"],
    value: Any,
    expected_bytes: bytes,
) -> None:
    """Test packing single values with different word orders."""
    s = WordOrderAwareStruct(format_str, word_order=word_order)
    result = s.pack(value)
    assert result == expected_bytes


@pytest.mark.parametrize(
    ("format_str", "word_order", "values", "expected_bytes"),
    [
        # Multiple single-register values
        (">HH", "big", (0x0102, 0x0304), b"\x01\x02\x03\x04"),
        (">HH", "little", (0x0102, 0x0304), b"\x01\x02\x03\x04"),
        # Mix of 1-register and 2-register values
        (">HI", "big", (0x0102, 0x09080706), b"\x01\x02\x09\x08\x07\x06"),
        (">HI", "little", (0x0102, 0x09080706), b"\x01\x02\x07\x06\x09\x08"),
        # Mix: H (1 reg) + I (2 reg) + Q (4 reg)
        (
            ">HIQ",
            "big",
            (0x0102, 0x09080706, 0x0A0B0C0D0E0F0102),
            b"\x01\x02\x09\x08\x07\x06\x0a\x0b\x0c\x0d\x0e\x0f\x01\x02",
        ),
        (
            ">HIQ",
            "little",
            (0x0102, 0x09080706, 0x0A0B0C0D0E0F0102),
            b"\x01\x02\x07\x06\x09\x08\x01\x02\x0e\x0f\x0c\x0d\x0a\x0b",
        ),
        # Multiple 2-register values
        (">II", "big", (0x01020304, 0x05060708), b"\x01\x02\x03\x04\x05\x06\x07\x08"),
        (">II", "little", (0x01020304, 0x05060708), b"\x03\x04\x01\x02\x07\x08\x05\x06"),
        # Array notation
        (">2H", "big", (0x0102, 0x0304), b"\x01\x02\x03\x04"),
        (">2I", "little", (0x01020304, 0x05060708), b"\x03\x04\x01\x02\x07\x08\x05\x06"),
    ],
)
def test_pack_multiple_values(
    format_str: str,
    word_order: Literal["little", "big"],
    values: tuple[Any, ...],
    expected_bytes: bytes,
) -> None:
    """Test packing multiple values with different combinations."""
    s = WordOrderAwareStruct(format_str, word_order=word_order)
    result = s.pack(*values)
    assert result == expected_bytes


@pytest.mark.parametrize(
    ("format_str", "word_order", "data", "expected_values"),
    [
        (">H", "big", b"\x01\x02", (0x0102,)),
        (">H", "little", b"\x01\x02", (0x0102,)),
        (">I", "big", b"\x0a\x0b\x0c\x0d", (0x0A0B0C0D,)),
        (">I", "little", b"\x0c\x0d\x0a\x0b", (0x0A0B0C0D,)),
        (">Q", "little", b"\x01\x02\x0e\x0f\x0c\x0d\x0a\x0b", (0x0A0B0C0D0E0F0102,)),
        (
            ">HIQ",
            "little",
            b"\x01\x02\x07\x06\x09\x08\x01\x02\x0e\x0f\x0c\x0d\x0a\x0b",
            (0x0102, 0x09080706, 0x0A0B0C0D0E0F0102),
        ),
    ],
)
def test_unpack(
    format_str: str,
    word_order: Literal["little", "big"],
    data: bytes,
    expected_values: tuple[Any, ...],
) -> None:
    """Test unpacking data with different word orders."""
    s = WordOrderAwareStruct(format_str, word_order=word_order)
    result = s.unpack(data)
    assert result == expected_values


@pytest.mark.parametrize(
    ("format_str", "word_order", "values"),
    [
        (">H", "big", (0x1234,)),
        (">I", "little", (0x12345678,)),
        (">Q", "little", (0x0123456789ABCDEF,)),
        (">HIQ", "little", (0x1234, 0x56789ABC, 0x0123456789ABCDEF)),
        (">2I", "little", (0x11111111, 0x22222222)),
        (">4H", "big", (0x1111, 0x2222, 0x3333, 0x4444)),
    ],
)
def test_roundtrip(
    format_str: str,
    word_order: Literal["little", "big"],
    values: tuple[Any, ...],
) -> None:
    """Test that pack/unpack roundtrips correctly."""
    s = WordOrderAwareStruct(format_str, word_order=word_order)
    packed = s.pack(*values)
    unpacked = s.unpack(packed)
    assert unpacked == values


def test_unpack_from() -> None:
    """Test unpack_from with offset."""
    s = WordOrderAwareStruct(">I", word_order="big")
    data = b"\x00\x00\x0a\x0b\x0c\x0d\x00\x00"
    result = s.unpack_from(data, offset=2)
    assert result == (0x0A0B0C0D,)


def test_pack_into() -> None:
    """Test pack_into with buffer and offset."""
    s = WordOrderAwareStruct(">I", word_order="little")
    buffer = bytearray(8)
    s.pack_into(buffer, 2, 0x0A0B0C0D)
    assert buffer == b"\x00\x00\x0c\x0d\x0a\x0b\x00\x00"


@pytest.mark.parametrize(
    ("format_str", "expected_lengths"),
    [
        (">H", [2]),
        (">I", [4]),
        (">Q", [8]),
        (">2H", [2, 2]),
        (">HIQ", [2, 4, 8]),
        (">3I", [4, 4, 4]),
        (">HH2I", [2, 2, 4, 4]),
        (">10s", [10]),
        (">11sbHH", [11, 1, 2, 2]),
    ],
)
def test_parse_format_lengths(format_str: str, expected_lengths: list[int]) -> None:
    """Test parsing format string into value lengths."""
    result = WordOrderAwareStruct.parse_format_lengths(format_str)
    assert result == expected_lengths


def test_parse_format_with_different_endianness() -> None:
    """Test that parse_format_lengths handles byte order characters."""
    for prefix in ["@", "=", "<", ">", "!", ""]:
        result = WordOrderAwareStruct.parse_format_lengths(f"{prefix}HIQ")
        assert result == [2, 4, 8]


def test_size_property() -> None:
    """Test that size property works correctly."""
    assert WordOrderAwareStruct(">H").size == 2
    assert WordOrderAwareStruct(">I").size == 4
    assert WordOrderAwareStruct(">HIQ").size == 14


def test_value_lengths_only_set_for_little_endian() -> None:
    """Test that _value_lengths is only populated for little endian word order."""
    big_struct = WordOrderAwareStruct(">I", word_order="big")
    assert big_struct._value_lengths is None

    little_struct = WordOrderAwareStruct(">I", word_order="little")
    assert little_struct._value_lengths == [4]


def test_odd_length_values_in_swap() -> None:
    """Test struct with odd-length individual values (like 'b' padded to even)."""
    # Test a struct with byte values that need padding to reach even length
    # Use 'BB' (2 bytes total) with little endian word order
    s = WordOrderAwareStruct(">BB", word_order="little")
    result = s.pack(0x12, 0x34)
    assert result == b"\x12\x34"  # No swapping for single register



"""Base class for Modbus PDU (Protocol Data Unit) handling."""

from abc import ABC, abstractmethod
from typing import Self, TypeVar

from tmodbus.exceptions import InvalidRequestError, InvalidResponseError

RT = TypeVar("RT")


class BaseClientPDU[RT](ABC):
    """Base class that defines the functions needed to handle Modbus PDUs on the client-side."""

    function_code: int
    rtu_response_data_length: int | None = None

    @abstractmethod
    def encode_request(self) -> bytes:
        """Convert PDU to bytes.

        This method should be implemented by subclasses to convert the PDU
        into a byte representation suitable for transmission.
        """

    @abstractmethod
    def decode_response(self, response: bytes) -> RT:
        """Decode the response PDU.

        Args:
            response: Response PDU bytes

        Returns:
            Decoded response data, type depends on the specific PDU implementation

        """

    @classmethod
    def get_expected_response_data_length(cls, data: bytes) -> int:
        """Get the expected number of bytes for the data part of the response PDU.

        This method should be implemented by subclasses to return the expected
        length of the response based on the specific PDU type.

        Returns:
            Expected length of the response PDU in bytes

        """
        # if a fixed length is defined for the response PDU, return it
        if cls.rtu_response_data_length is not None:
            return cls.rtu_response_data_length

        # otherwise, we assume that the first byte of the PDU-part of the response denotes
        # the total length of the PDU.
        # If this is not the case (ex. for function code 0x18), the subclass should override this method.
        return (
            1  # the first byte containing the total length of the PDU
            + data[0]
        )


class BasePDU[RT](BaseClientPDU[RT]):
    """Base class that defines the functions needed to handle Modbus PDUs on both the client-side and server-side."""

    rtu_request_data_length: int | None = None

    ### Server methods ###

    @classmethod
    @abstractmethod
    def decode_request(cls, request: bytes) -> Self:
        """Create an instance of this PDU from a request byte sequence."""

    @abstractmethod
    def encode_response(self, value: RT) -> bytes:
        """Convert the response value to bytes.

        Args:
            value: The value to encode in the response

        Returns:
            Bytes representation of the response PDU

        """

    @classmethod
    def get_expected_request_data_length(cls, data: bytes) -> int:
        """Get the expected number of bytes for the data part of the request PDU.

        This method should be implemented by subclasses to return the expected
        length of the request based on the specific PDU type.

        Returns:
            Expected length of the request PDU in bytes

        """
        # if a fixed length is defined for the request PDU, return it
        if cls.rtu_request_data_length is not None:
            return cls.rtu_request_data_length

        # otherwise, we assume that the first byte of the PDU-part of the response denotes
        # the total length of the PDU.
        # If this is not the case (ex. for function code 0x18), the subclass should override this method.
        return (
            1  # the first byte containing the total length of the PDU
            + data[0]
        )


class BaseSubFunctionClientPDU[RT](BaseClientPDU[RT]):
    """Extends the BaseClientPDU to include sub-function code.

    Only the get_expected_response_data_length method is changed in this class.
    """

    sub_function_code: int

    @classmethod
    def get_expected_response_data_length(cls, data: bytes) -> int:
        """Get the expected number of bytes for the data part of the response PDU.

        This method should be implemented by subclasses to return the expected
        length of the response based on the specific PDU type.

        Returns:
            Expected length of the response PDU in bytes

        """
        # Always assume that the first byte of the data-part of the frame contains the sub-function code
        if data[0] != cls.sub_function_code:
            msg = f"Expected sub-function code {cls.sub_function_code}, got {data[0]}"
            raise InvalidResponseError(msg, response_bytes=data)

        # if a fixed length is defined for the response PDU, return it
        if cls.rtu_response_data_length is not None:
            return cls.rtu_response_data_length

        # otherwise, we assume that the second byte of the data-part of the frame contains the total length of the PDU.
        return (
            1  # the first byte containing the sub-function code
            + 1  # the second byte containing the total length of the PDU
            + data[1]
        )


class BaseSubFunctionPDU[RT](BaseSubFunctionClientPDU[RT], BasePDU[RT]):
    """Extends the BaseServerPDU to include sub-function code.

    Only the get_expected_response_data_length method is changed in this class.
    """

    sub_function_code: int

    @classmethod
    def get_expected_response_data_length(cls, data: bytes) -> int:
        """Get the expected number of bytes for the data part of the response PDU.

        This method should be implemented by subclasses to return the expected
        length of the response based on the specific PDU type.

        Returns:
            Expected length of the response PDU in bytes

        """
        # Always assume that the first byte of the data-part of the frame contains the sub-function code
        if data[0] != cls.sub_function_code:
            msg = f"Expected sub-function code {cls.sub_function_code}, got {data[0]}"
            raise InvalidResponseError(msg, response_bytes=data)

        # if a fixed length is defined for the response PDU, return it
        if cls.rtu_response_data_length is not None:
            return cls.rtu_response_data_length

        # otherwise, we assume that the second byte of the data-part of the frame contains the total length of the PDU.
        return (
            1  # the first byte containing the sub-function code
            + 1  # the second byte containing the total length of the PDU
            + data[1]
        )

    @classmethod
    def get_expected_request_data_length(cls, data: bytes) -> int:
        """Get the expected number of bytes for the data part of the request PDU.

        This method should be implemented by subclasses to return the expected
        length of the request based on the specific PDU type.

        Returns:
            Expected length of the request PDU in bytes

        """
        # Always assume that the first byte of the data-part of the frame contains the sub-function code
        if data[0] != cls.sub_function_code:
            msg = f"Expected sub-function code {cls.sub_function_code}, got {data[0]}"
            raise InvalidRequestError(msg, request_bytes=data)

        # if a fixed length is defined for the request PDU, return it
        if cls.rtu_request_data_length is not None:
            return cls.rtu_request_data_length

        # otherwise, we assume that the first byte of the PDU-part of the response denotes
        # the total length of the PDU.
        # If this is not the case (ex. for function code 0x18), the subclass should override this method.
        return (
            1  # the first byte containing the total length of the PDU
            + data[0]
        )



"""Modbus Protocol Data Unit (PDU)."""

from typing import Any

from tmodbus.const import FunctionCode

from .base import BaseClientPDU, BasePDU, BaseSubFunctionClientPDU, BaseSubFunctionPDU
from .coils import ReadCoilsPDU, WriteMultipleCoilsPDU, WriteSingleCoilPDU
from .device import ReadDeviceIdentificationPDU, ReadDeviceIdentificationResponse
from .discrete_inputs import ReadDiscreteInputsPDU
from .holding_registers import (
    ReadHoldingRegistersPDU,
    ReadInputRegistersPDU,
    WriteMultipleRegistersPDU,
    WriteSingleRegisterPDU,
)

function_code_to_pdu_map: dict[int, type[BaseClientPDU[Any]]] = {
    FunctionCode.READ_COILS: ReadCoilsPDU,
    FunctionCode.READ_DISCRETE_INPUTS: ReadDiscreteInputsPDU,
    FunctionCode.READ_HOLDING_REGISTERS: ReadHoldingRegistersPDU,
    FunctionCode.READ_INPUT_REGISTERS: ReadInputRegistersPDU,
    FunctionCode.WRITE_SINGLE_COIL: WriteSingleCoilPDU,
    FunctionCode.WRITE_SINGLE_REGISTER: WriteSingleRegisterPDU,
    FunctionCode.WRITE_MULTIPLE_COILS: WriteMultipleCoilsPDU,
    FunctionCode.WRITE_MULTIPLE_REGISTERS: WriteMultipleRegistersPDU,
}

sub_function_code_to_pdu_map: dict[int, dict[int, type[BaseSubFunctionClientPDU[Any]]]] = {
    FunctionCode.ENCAPSULATED_INTERFACE_TRANSPORT: {
        ReadDeviceIdentificationPDU.sub_function_code: ReadDeviceIdentificationPDU,
    }
}


def register_pdu_class(pdu_class: type[BaseClientPDU[Any]]) -> None:
    """Register a PDU class for a specific function code.

    Args:
        pdu_class: PDU class to register

    """
    function_code = pdu_class.function_code
    if issubclass(pdu_class, BaseSubFunctionClientPDU):
        if existing_pdu_class := function_code_to_pdu_map.get(function_code):
            msg = (
                f"Function code {function_code:#04x} is already registered "
                f"for a non-subfunction PDU {existing_pdu_class.__name__}."
            )
            raise ValueError(msg)
        if function_code not in sub_function_code_to_pdu_map:
            sub_function_code_to_pdu_map[function_code] = {}

        sub_function_code = pdu_class.sub_function_code

        if sub_function_code in sub_function_code_to_pdu_map[function_code]:
            msg = (
                f"A PDU with function code {function_code:#04x}, "
                f"and sub-function code {sub_function_code:#04x} is already registered: "
                f"{sub_function_code_to_pdu_map[function_code][sub_function_code].__name__}."
            )
            raise ValueError(msg)

        sub_function_code_to_pdu_map[function_code][sub_function_code] = pdu_class
    else:  # Registering a normal PDU class
        if existing_sub_pdus := sub_function_code_to_pdu_map.get(function_code):
            existing_sub_pdus_list = ", ".join(
                f"{sub_function_code:#04x}: {c.__name__}" for sub_function_code, c in existing_sub_pdus.items()
            )
            msg = (
                f"Function code {function_code:#04x} is already registered with sub-functions: {existing_sub_pdus_list}"
            )
            raise ValueError(msg)

        if existing_pdu_class := function_code_to_pdu_map.get(function_code):
            msg = f"Function code {function_code:#02x} is already registered to {existing_pdu_class.__name__}."
            raise ValueError(msg)

        function_code_to_pdu_map[pdu_class.function_code] = pdu_class


def get_pdu_class(begin_bytes: bytes) -> type[BaseClientPDU[Any]]:
    """Get PDU class by function code.

    Args:
        begin_bytes: Beginning bytes of the PDU, starting with function code

    Returns:
        Corresponding PDU class

    Raises:
        ValueError: If function code is not supported

    """
    function_code = begin_bytes[0]
    if pdu_class := function_code_to_pdu_map.get(function_code):
        return pdu_class

    if sub_function_pdus := sub_function_code_to_pdu_map.get(function_code):
        sub_function_code = begin_bytes[1]
        try:
            return sub_function_pdus[sub_function_code]
        except KeyError:
            msg = f"Unsupported sub-function code: {sub_function_code:#02x} for function code {function_code:#02x}"
            raise ValueError(msg) from None

    msg = f"Unsupported function code: {function_code:#02x}"
    raise ValueError(msg)


__all__ = [
    "BaseClientPDU",
    "BasePDU",
    "BaseSubFunctionClientPDU",
    "BaseSubFunctionPDU",
    "ReadCoilsPDU",
    "ReadDeviceIdentificationPDU",
    "ReadDeviceIdentificationResponse",
    "ReadDiscreteInputsPDU",
    "ReadHoldingRegistersPDU",
    "ReadInputRegistersPDU",
    "WriteMultipleCoilsPDU",
    "WriteMultipleRegistersPDU",
    "WriteSingleCoilPDU",
    "WriteSingleRegisterPDU",
    "register_pdu_class",
]



"""Exceptions."""

from typing import Any

from .const import ExceptionCode


class TModbusError(Exception):
    """Base exception class for ModbusLink library."""


class ModbusConnectionError(TModbusError):
    """Connection error exception.

    Raised when unable to establish or maintain connection with Modbus device.
    """

    response_bytes: bytes
    """The bytes that were read before the connection error occurred. Can be empty."""

    def __init__(self, *args: Any, bytes_read: bytes | None = None, **kwargs: Any) -> None:
        """Initialize RTUFrameError."""
        super().__init__(*args, **kwargs)
        self.response_bytes = bytes_read or b""


class InvalidRequestError(TModbusError):
    """Invalid request error exception.

    Raised when a request is malformed or invalid.
    """

    def __init__(self, *args: Any, request_bytes: bytes | None = None, **kwargs: Any) -> None:  # pragma: no cover
        """Initialize InvalidRequestError."""
        super().__init__(*args, **kwargs)
        self.request_bytes = request_bytes or b""


class InvalidResponseError(TModbusError):
    """Invalid response error exception.

    Raised when received response format is incorrect or unexpected.
    """

    response_bytes: bytes

    def __init__(self, *args: Any, response_bytes: bytes, **kwargs: Any) -> None:
        """Initialize RTUFrameError."""
        super().__init__(*args, **kwargs)
        self.response_bytes = response_bytes


class RTUFrameError(InvalidResponseError):
    """RTU frame error exception.

    Raised when there is a framing error in the received RTU frame.
    """


class CRCError(InvalidResponseError):
    """CRC validation error exception.

    Raised when CRC validation of received data frame fails.
    """


class HeaderMismatchError(InvalidResponseError):
    """Header mismatch error exception.

    Raised when the header of the received response does not match the request.
    """


class FunctionCodeError(InvalidResponseError):
    """Function code error exception.

    Raised when the function code in the response does not match the request.
    """


class RequestRetryFailedError(TModbusError):
    """Failed to get an appropriate response after exhausting all retries.

    Raised when all retry attempts for a request have been exhausted without success.
    """


class ModbusResponseError(TModbusError):
    """Base class for all Modbus exception response."""

    error_code: int

    def __init__(self, error_code: int, function_code: int) -> None:
        """Initialize ModbusResponseError.

        Args:
            error_code: Error code from the Modbus exception response
            function_code: Function code of the request that caused the exception

        """
        super().__init__(f"Modbus Exception {error_code:#04x} for function code 0x{function_code:#04x}")
        assert self.error_code == error_code
        self.function_code = function_code


class IllegalFunctionError(ModbusResponseError):
    """The function code received in the request is not an allowable action for the server."""

    error_code = ExceptionCode.ILLEGAL_FUNCTION


class IllegalDataAddressError(ModbusResponseError):
    """The data address received in the request is not an allowable address for the server."""

    error_code = ExceptionCode.ILLEGAL_DATA_ADDRESS


class IllegalDataValueError(ModbusResponseError):
    """The value contained in the request data field is not an allowable value for the server."""

    error_code = ExceptionCode.ILLEGAL_DATA_VALUE


class ServerDeviceFailureError(ModbusResponseError):
    """An unrecoverable error occurred."""

    error_code = ExceptionCode.SERVER_DEVICE_FAILURE


class AcknowledgeError(ModbusResponseError):
    """Acknowledge error.

    The server has accepted the requests and it processing it,
    but a long duration of time will be required to do so.
    """

    error_code = ExceptionCode.ACKNOWLEDGE


class ServerDeviceBusyError(ModbusResponseError):
    """The server is engaged in a long-duration program command."""

    error_code = ExceptionCode.SERVER_DEVICE_BUSY


class MemoryParityError(ModbusResponseError):
    """The server attempted to read record file, but detected a parity error in memory."""

    error_code = ExceptionCode.MEMORY_PARITY_ERROR


class GatewayPathUnavailableError(ModbusResponseError):
    """The gateway is probably misconfigured or overloaded."""

    error_code = ExceptionCode.GATEWAY_PATH_UNAVAILABLE


class GatewayTargetDeviceFailedToRespondError(ModbusResponseError):
    """Didn't get a response from target device."""

    error_code = ExceptionCode.GATEWAY_TARGET_DEVICE_FAILED_TO_RESPOND


class AbnormalDeviceDescriptionError(ModbusResponseError):
    """The device description definition call returned a response."""

    error_code = ExceptionCode.ABNORNMAL_DEVICE_DESCRIPTION


error_code_to_exception_map: dict[int, type[ModbusResponseError]] = {
    IllegalFunctionError.error_code: IllegalFunctionError,
    IllegalDataAddressError.error_code: IllegalDataAddressError,
    IllegalDataValueError.error_code: IllegalDataValueError,
    ServerDeviceFailureError.error_code: ServerDeviceFailureError,
    AcknowledgeError.error_code: AcknowledgeError,
    ServerDeviceBusyError.error_code: ServerDeviceBusyError,
    MemoryParityError.error_code: MemoryParityError,
    GatewayPathUnavailableError.error_code: GatewayPathUnavailableError,
    GatewayTargetDeviceFailedToRespondError.error_code: GatewayTargetDeviceFailedToRespondError,
    AbnormalDeviceDescriptionError.error_code: AbnormalDeviceDescriptionError,
}


def register_custom_exception(err_cls: type[ModbusResponseError]) -> None:
    """Register a custom Modbus exception class.

    Args:
        err_cls: Custom exception class to register

    """
    if err_cls.error_code in error_code_to_exception_map:
        msg = f"Error code {err_cls.error_code} is already registered."
        raise ValueError(msg)

    error_code_to_exception_map[err_cls.error_code] = err_cls



"""Async Smart Transport Layer.

This transport layer lives on top of the RTU or TCP transport layer
and implements the following smart features:

- Wait time between requests to avoid overwhelming the device
- Wait time after connection establishment to allow device to be ready
- Automatic reconnection on connection loss
"""

import asyncio
import logging
import time
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, TypeVar

try:
    from tenacity import (
        AsyncRetrying,
        RetryError,
        retry_any,
        retry_if_exception_type,
        stop_after_delay,
        wait_exponential,
    )
except ImportError as ex:  # pragma: no cover
    msg = "tenacity is required for Smart Transport functionality.Install with 'pip install tmodbus[smart]'"
    raise ImportError(msg) from ex

from tmodbus.exceptions import (
    ModbusConnectionError,
    RequestRetryFailedError,
    ServerDeviceBusyError,
    ServerDeviceFailureError,
)
from tmodbus.pdu import BaseClientPDU

from .async_base import AsyncBaseTransport

if TYPE_CHECKING:
    from tenacity import retry_base

logger = logging.getLogger(__name__)


RT = TypeVar("RT")


DEFAULT_RECONNECT_RETRY_STRATEGY = AsyncRetrying(
    stop=stop_after_delay(60),
    wait=wait_exponential(min=0.1, max=10),
)

DEFAULT_RESPONSE_RETRY_STRATEGY = AsyncRetrying(
    stop=stop_after_delay(60),
    wait=wait_exponential(min=0.1, max=10),
    reraise=True,  # Reraise the last exception if all retries are exhausted
)


class AsyncSmartTransport(AsyncBaseTransport):
    """Smart Transport Layer.

    This transport layer is built on top of RTU or TCP transport layers,
    adding features such as wait time between requests, wait time after connection,
    and automatic reconnection on connection loss.
    """

    _communication_lock: asyncio.Lock = asyncio.Lock()
    _should_be_connected: bool = False

    auto_reconnect: AsyncRetrying | None = None
    response_retry_strategy: AsyncRetrying

    def __init__(  # noqa: PLR0913
        self,
        base_transport: "AsyncBaseTransport",
        *,
        wait_between_requests: float = 0.0,
        wait_after_connect: float = 0.0,
        auto_reconnect: bool | AsyncRetrying = True,
        on_reconnected: Callable[[], Awaitable[None] | None] | None = None,
        response_retry_strategy: AsyncRetrying | None = None,
        retry_on_device_busy: bool = True,
        retry_on_device_failure: bool = False,
    ) -> None:
        """Initialize Smart Transport Layer.

        Args:
            base_transport: Underlying transport layer instance (AsyncRtuTransport, AsyncTcpTransport, etc.)
            wait_between_requests: Wait time between requests in seconds (default: 0.0s)
            wait_after_connect: Wait time after connection establishment in seconds (default: 0.0s)
            auto_reconnect: Whether to automatically reconnect on connection loss (default: True).
                            Can be a custom AsyncRetrying instance when more control is needed.
            on_reconnected: Callback to be called after a successful reconnection.
            response_retry_strategy: Retry strategy for handling failed requests (default: None).
            retry_on_device_busy: Whether to retry on device busy errors (default: True).
                                  Can be a custom AsyncRetrying instance when more control is needed.
            retry_on_device_failure: Whether to retry on device failure errors (default: False).
                                     Can be a custom AsyncRetrying instance when more control is needed.

        """
        self.base_transport = base_transport
        if wait_between_requests < 0:
            msg = "wait_between_requests must be a positive value"
            raise ValueError(msg)
        self.wait_between_requests = wait_between_requests

        if wait_after_connect < 0:
            msg = "wait_after_connect must be a positive value"
            raise ValueError(msg)
        self.wait_after_connect = wait_after_connect

        if isinstance(auto_reconnect, bool) and auto_reconnect:
            auto_reconnect = DEFAULT_RECONNECT_RETRY_STRATEGY
        if auto_reconnect:
            self.auto_reconnect = auto_reconnect.copy(
                retry=retry_if_exception_type((ModbusConnectionError, TimeoutError))
            )

        if not auto_reconnect and on_reconnected:
            msg = "on_reconnected callback provided but auto_reconnect is disabled"
            raise ValueError(msg)
        self.on_reconnected = on_reconnected

        retry_functions: list[retry_base] = [
            # retry if connection was interrupted
            retry_if_exception_type((ConnectionResetError, asyncio.IncompleteReadError)),
        ]
        if not response_retry_strategy:
            response_retry_strategy = DEFAULT_RESPONSE_RETRY_STRATEGY
        elif response_retry_strategy.retry:
            retry_functions.append(response_retry_strategy.retry)  # type: ignore[arg-type]

        if retry_on_device_busy:
            retry_functions.append(retry_if_exception_type(ServerDeviceBusyError))

        if retry_on_device_failure:
            retry_functions.append(retry_if_exception_type(ServerDeviceFailureError))

        self.response_retry_strategy = response_retry_strategy.copy(retry=retry_any(*retry_functions))

        self._last_request_finished_at: float | None = None

    async def open(self) -> None:
        """Open Transport Connection.

        Establishes connection with Modbus device and waits for the specified time
        to allow the device to be ready.

        Raises:
            ConnectionError: When connection cannot be established

        """
        async with self._communication_lock:
            await self.base_transport.open()
            if self.wait_after_connect > 0:
                logger.debug("Waiting %.2f seconds after TCP connection before sending data", self.wait_after_connect)
                await asyncio.sleep(self.wait_after_connect)

            self._should_be_connected = True

    async def close(self) -> None:
        """Close Transport Connection.

        Closes connection with Modbus device and releases related resources.
        """
        async with self._communication_lock:
            try:
                await self.base_transport.close()
            finally:
                self._should_be_connected = False

    def is_open(self) -> bool:
        """Check Connection Status.

        Returns:
            True if connection was established and should still be available.
            False otherwise.

        """
        return (self.auto_reconnect and self._should_be_connected) or self.base_transport.is_open()

    async def _do_auto_reconnect(self) -> None:
        """Reconnect to the Modbus device."""
        assert isinstance(self.auto_reconnect, AsyncRetrying)

        try:
            async for attempt in self.auto_reconnect:
                with attempt:
                    logger.info("Attempting to reconnect.")
                    await self.open()
        except RetryError as e:
            msg = (
                f"Failed to reconnect after {attempt.retry_state.attempt_number} attempts "
                f"over {attempt.retry_state.seconds_since_start} seconds"
            )
            raise ModbusConnectionError(msg) from e

        if self.on_reconnected:
            result = self.on_reconnected()
            if asyncio.iscoroutine(result):
                await result

    async def _reconnect_send_and_receive(self, unit_id: int, pdu: BaseClientPDU[RT]) -> RT:
        """Reconnect if necessary, then try to Send PDU and Receive Response."""
        # If auto_reconnect is enabled and the connection is not open, try to reconnect
        if self.auto_reconnect and not self.base_transport.is_open():
            logger.info("Connection lost. Attempting to reconnect...")
            await self._do_auto_reconnect()

        # If a wait time between requests is configured, enforce it
        if self.wait_between_requests > 0 and self._last_request_finished_at is not None:
            wait_needed = self.wait_between_requests - (time.monotonic() - self._last_request_finished_at)
            if wait_needed > 0:
                logger.debug(
                    "Waiting %.2fs before sending next request to respect %.2fs wait between requests",
                    wait_needed,
                    self.wait_between_requests,
                )
                await asyncio.sleep(wait_needed)

        return await self.base_transport.send_and_receive(unit_id, pdu)

    async def send_and_receive(self, unit_id: int, pdu: BaseClientPDU[RT]) -> RT:
        """Send PDU and Receive Response."""
        # Ensure that only one request is processed at a time
        async with self._communication_lock:
            try:
                async for attempt in self.response_retry_strategy:
                    with attempt:
                        response = await self._reconnect_send_and_receive(unit_id, pdu)

                    if attempt.retry_state.outcome and not attempt.retry_state.outcome.failed:
                        attempt.retry_state.set_result(response)
            except RetryError as e:
                msg = (
                    f"Failed to get a valid response after {attempt.retry_state.attempt_number} attempts "
                    f"over {attempt.retry_state.seconds_since_start} seconds"
                )
                raise RequestRetryFailedError(msg) from e
            else:
                return response
            finally:
                self._last_request_finished_at = time.monotonic()



"""Tests for tmodbus/transport/async_smart.py ."""

import time
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from tenacity import AsyncRetrying, retry_if_exception_type, stop_after_attempt
from tmodbus.exceptions import ModbusConnectionError, RequestRetryFailedError
from tmodbus.pdu.base import BaseClientPDU
from tmodbus.transport.async_base import AsyncBaseTransport
from tmodbus.transport.async_smart import AsyncSmartTransport


class DummyPDU(BaseClientPDU[tuple[str, bytes]]):
    """Dummy PDU for testing."""

    function_code = 0x03

    def encode_request(self) -> bytes:
        """Encode request."""
        return b"\x03\x00"

    def decode_response(self, data: bytes) -> tuple[str, bytes]:
        """Decode response."""
        return ("ok", data)


@pytest.fixture
def base_transport_mock() -> AsyncBaseTransport:
    """Fixture to create a mock AsyncBaseTransport."""
    bt = MagicMock(spec=AsyncBaseTransport)
    bt.open = AsyncMock()
    bt.close = AsyncMock()
    bt.is_open = lambda: True
    bt.send_and_receive = AsyncMock(return_value=("ok", b""))
    return bt


def test_init_negative_waits(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that negative wait times raise ValueError."""
    with pytest.raises(ValueError, match=r"wait_between_requests must be .*"):
        AsyncSmartTransport(base_transport_mock, wait_between_requests=-1)
    with pytest.raises(ValueError, match=r"wait_after_connect must be .*"):
        AsyncSmartTransport(base_transport_mock, wait_after_connect=-0.1)


def test_on_reconnected_requires_auto_reconnect(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that providing on_reconnected without auto_reconnect raises ValueError."""
    with pytest.raises(ValueError, match="on_reconnected callback provided but auto_reconnect is disabled"):
        AsyncSmartTransport(base_transport_mock, auto_reconnect=False, on_reconnected=lambda: None)


async def test_open_waits_after_connect(base_transport_mock: MagicMock) -> None:
    """Test that open waits after connecting if configured."""
    t = AsyncSmartTransport(base_transport_mock, wait_after_connect=0.05)

    with patch("asyncio.sleep", AsyncMock()) as fake_sleep:
        await t.open()
        base_transport_mock.open.assert_awaited()
        fake_sleep.assert_awaited()
    assert t._should_be_connected


async def test_close_resets_should_be_connected(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that close resets _should_be_connected flag."""
    t = AsyncSmartTransport(base_transport_mock, wait_after_connect=0.0)
    await t.open()
    assert t._should_be_connected
    await t.close()
    assert not t._should_be_connected


async def test_reconnect_and_wait_between_requests(base_transport_mock: MagicMock) -> None:
    """Test that _reconnect_send_and_receive waits and reconnects as needed."""
    # make base_transport initially closed
    base_transport_mock.is_open = lambda: False

    t = AsyncSmartTransport(base_transport_mock, wait_between_requests=0.1)

    # stub _do_auto_reconnect to simulate reconnection (set is_open True)
    async def do_reconnect() -> None:
        base_transport_mock.is_open = lambda: True

    # ensure last request finished just now to trigger wait
    t._last_request_finished_at = time.monotonic()

    do_auto_reconnect_mock = AsyncMock(side_effect=do_reconnect)

    with (
        patch.object(t, "_do_auto_reconnect", do_auto_reconnect_mock),
        patch("asyncio.sleep", AsyncMock()) as fake_sleep,
    ):
        resp = await t._reconnect_send_and_receive(1, DummyPDU())
        do_auto_reconnect_mock.assert_awaited()
        fake_sleep.assert_awaited()
        base_transport_mock.send_and_receive.assert_awaited()
        assert resp == ("ok", b"")


async def test_send_and_receive_updates_last_finished(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that last_request_finished_at is updated after send_and_receive."""
    t = AsyncSmartTransport(base_transport_mock)

    # stub the underlying send to return quickly
    with patch.object(t, "_reconnect_send_and_receive", AsyncMock(return_value=("ok", b""))):
        before = time.monotonic()
        resp = await t.send_and_receive(1, DummyPDU())
        after = time.monotonic()

        assert resp == ("ok", b"")
        assert t._last_request_finished_at is not None
        assert before <= t._last_request_finished_at <= after


async def test_do_auto_reconnect_retry_error(base_transport_mock: MagicMock) -> None:
    """Test that ModbusConnectionError is raised when auto_reconnect exhausts attempts."""
    t = AsyncSmartTransport(
        base_transport_mock, auto_reconnect=AsyncRetrying(stop=stop_after_attempt(1), reraise=False)
    )
    base_transport_mock.open.side_effect = ModbusConnectionError("fail")

    with pytest.raises(ModbusConnectionError):
        await t._do_auto_reconnect()


async def test_do_auto_reconnect_calls_on_reconnected(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test that on_reconnected callback is called after successful reconnection."""
    on_reconnected = AsyncMock()

    t = AsyncSmartTransport(
        base_transport_mock,
        on_reconnected=on_reconnected,
        auto_reconnect=AsyncRetrying(stop=stop_after_attempt(1), reraise=True),
    )

    await t._do_auto_reconnect()
    on_reconnected.assert_called_once()


async def test_send_and_receive_request_retry_failed(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that ModbusConnectionError is raised when response_retry_strategy exhausts attempts."""
    t = AsyncSmartTransport(base_transport_mock)

    # make _reconnect_send_and_receive raise each attempt; response_retry_strategy raises RetryError

    t.response_retry_strategy = AsyncRetrying(stop=stop_after_attempt(1), reraise=True)

    with (
        patch.object(t, "_reconnect_send_and_receive", AsyncMock(side_effect=ModbusConnectionError("fail"))),
        pytest.raises(ModbusConnectionError),
    ):
        await t.send_and_receive(1, DummyPDU())
    assert t._last_request_finished_at is not None


async def test_send_and_receive_response_retry_success(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test that send_and_receive succeeds when response_retry_strategy yields a successful attempt."""
    t = AsyncSmartTransport(
        base_transport_mock,
        response_retry_strategy=AsyncRetrying(stop=stop_after_attempt(1), reraise=True),
        retry_on_device_failure=True,
    )
    # success path: response_retry_strategy yields an attempt whose retry_state.outcome.failed == False
    with patch.object(t, "_reconnect_send_and_receive", AsyncMock(return_value=("ok", b""))):
        resp = await t.send_and_receive(1, DummyPDU())
        assert resp == ("ok", b"")
        assert t._last_request_finished_at is not None


async def test_send_and_receive_retry_strategy_raises_request_retry_failed(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test that RequestRetryFailedError is raised when response_retry_strategy exhausts attempts."""
    t = AsyncSmartTransport(
        base_transport_mock,
        # configure response_retry_strategy that yields at least one attempt then raises RetryError
        response_retry_strategy=AsyncRetrying(stop=stop_after_attempt(1), reraise=False),
    )
    with patch.object(t, "_reconnect_send_and_receive", AsyncMock(side_effect=ModbusConnectionError("fail"))):
        with pytest.raises(RequestRetryFailedError):
            await t.send_and_receive(1, DummyPDU())
        assert t._last_request_finished_at is not None


def test_is_open_cases() -> None:
    """Test is_open method behavior under different conditions."""
    bt = MagicMock()
    bt.is_open = lambda: False
    t = AsyncSmartTransport(bt)
    # simulate should be connected and auto_reconnect present
    t._should_be_connected = True
    assert t.is_open()

    t2 = AsyncSmartTransport(bt, auto_reconnect=False)
    bt.is_open = lambda: False
    assert not t2.is_open()


async def test_send_and_receive_request_retry_failed_raises_and_sets_timestamp(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test that when _reconnect_send_and_receive raises, the timestamp is still set."""
    t = AsyncSmartTransport(
        base_transport_mock,
        # set strategy that will raise RetryError after attempts
        response_retry_strategy=AsyncRetrying(stop=stop_after_attempt(1), reraise=True),
    )

    with patch.object(t, "_reconnect_send_and_receive", AsyncMock(side_effect=ConnectionResetError("boom"))):
        with pytest.raises(ConnectionResetError):
            await t.send_and_receive(1, DummyPDU())
        assert t._last_request_finished_at is not None


async def test_send_and_receive_else_branch_sets_timestamp(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that the else branch sets the last_request_finished_at timestamp."""
    # ensure no response retry strategy
    t = AsyncSmartTransport(base_transport_mock, response_retry_strategy=None, auto_reconnect=False)
    with patch.object(t, "_reconnect_send_and_receive", AsyncMock(return_value=("ok", b""))):
        resp = await t.send_and_receive(1, DummyPDU())
        assert resp == ("ok", b"")
        assert t._last_request_finished_at is not None


async def test_do_auto_reconnect_without_on_reconnected(base_transport_mock: MagicMock) -> None:
    """Test that _do_auto_reconnect succeeds without on_reconnected callback."""
    t = AsyncSmartTransport(
        base_transport_mock,
        auto_reconnect=AsyncRetrying(stop=stop_after_attempt(1), reraise=True),
    )

    # Should not raise, and should not call on_reconnected since it's None
    await t._do_auto_reconnect()
    base_transport_mock.open.assert_awaited()


async def test_do_auto_reconnect_with_sync_on_reconnected(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that on_reconnected callback is called when it's a sync function."""
    on_reconnected_called = False

    def sync_on_reconnected() -> None:
        nonlocal on_reconnected_called
        on_reconnected_called = True

    t = AsyncSmartTransport(
        base_transport_mock,
        on_reconnected=sync_on_reconnected,
        auto_reconnect=AsyncRetrying(stop=stop_after_attempt(1), reraise=True),
    )

    await t._do_auto_reconnect()
    assert on_reconnected_called


async def test_response_retry_strategy_with_custom_retry(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that custom response_retry_strategy with retry attribute is used."""
    custom_strategy = AsyncRetrying(
        stop=stop_after_attempt(2),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    )

    t = AsyncSmartTransport(
        base_transport_mock,
        response_retry_strategy=custom_strategy,
    )

    # The retry strategy should include the custom retry function
    assert t.response_retry_strategy is not None


async def test_response_retry_strategy_without_retry_attribute(base_transport_mock: AsyncBaseTransport) -> None:
    """Test that response_retry_strategy without retry attribute is handled."""
    custom_strategy = AsyncRetrying(
        stop=stop_after_attempt(2),
        reraise=True,
    )

    t = AsyncSmartTransport(
        base_transport_mock,
        response_retry_strategy=custom_strategy,
    )

    # Should not raise and should create a valid retry strategy
    assert t.response_retry_strategy is not None


async def test_reconnect_send_and_receive_without_auto_reconnect(base_transport_mock: MagicMock) -> None:
    """Test _reconnect_send_and_receive when auto_reconnect is disabled."""
    t = AsyncSmartTransport(base_transport_mock, auto_reconnect=False)

    # Should not attempt to reconnect even if connection is closed
    base_transport_mock.is_open = lambda: False

    # Should just call send_and_receive directly
    resp = await t._reconnect_send_and_receive(1, DummyPDU())
    assert resp == ("ok", b"")
    base_transport_mock.send_and_receive.assert_awaited()


async def test_reconnect_send_and_receive_without_wait_between_requests(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test _reconnect_send_and_receive when wait_between_requests is 0."""
    t = AsyncSmartTransport(base_transport_mock, wait_between_requests=0.0, auto_reconnect=False)

    # Set last request time but wait_between_requests is 0, so no wait
    t._last_request_finished_at = time.monotonic()

    with patch("asyncio.sleep", AsyncMock()) as fake_sleep:
        resp = await t._reconnect_send_and_receive(1, DummyPDU())
        assert resp == ("ok", b"")
        # Should not sleep since wait_between_requests is 0
        fake_sleep.assert_not_awaited()


async def test_reconnect_send_and_receive_with_negative_wait_needed(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test _reconnect_send_and_receive when wait_needed is negative (enough time has passed)."""
    t = AsyncSmartTransport(base_transport_mock, wait_between_requests=0.1, auto_reconnect=False)

    # Set last request time far in the past so wait_needed will be negative
    t._last_request_finished_at = time.monotonic() - 1.0

    with patch("asyncio.sleep", AsyncMock()) as fake_sleep:
        resp = await t._reconnect_send_and_receive(1, DummyPDU())
        assert resp == ("ok", b"")
        # Should not sleep since enough time has passed
        fake_sleep.assert_not_awaited()


async def test_retry_on_device_busy_disabled(base_transport_mock: AsyncBaseTransport) -> None:
    """Test initialization when retry_on_device_busy is False."""
    t = AsyncSmartTransport(
        base_transport_mock,
        retry_on_device_busy=False,
    )

    # Should still have a valid retry strategy
    assert t.response_retry_strategy is not None


async def test_response_retry_strategy_with_falsy_retry_attribute(
    base_transport_mock: AsyncBaseTransport,
) -> None:
    """Test response_retry_strategy where retry attribute evaluates to False."""
    # Create a custom strategy without a retry parameter
    # When AsyncRetrying is created without retry, it defaults to a value that could be falsy
    custom_strategy = AsyncRetrying(
        stop=stop_after_attempt(2),
        reraise=True,
        # Not specifying 'retry' parameter means it will have a default/empty retry
    )
    # Explicitly set to None to simulate falsy
    custom_strategy.retry = None  # type: ignore[assignment]

    # The strategy should work even without a custom retry
    t = AsyncSmartTransport(
        base_transport_mock,
        response_retry_strategy=custom_strategy,
    )

    # Should not raise and should create a valid retry strategy
    assert t.response_retry_strategy is not None



"""tModbus library."""

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Unpack

from .client.async_client import AsyncModbusClient
from .transport import AsyncRtuTransport, AsyncSmartTransport, AsyncTcpTransport
from .transport.async_rtu import PySerialOptions

if TYPE_CHECKING:
    from tenacity import AsyncRetrying

try:
    from ._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.0"


def create_async_tcp_client(  # noqa: PLR0913
    host: str,
    port: int = 502,
    *,
    unit_id: int,
    timeout: float = 10.0,
    connect_timeout: float = 10.0,
    wait_between_requests: float = 0.0,
    wait_after_connect: float = 0.0,
    auto_reconnect: "bool | AsyncRetrying" = True,
    on_reconnected: Callable[[], Awaitable[None] | None] | None = None,
    response_retry_strategy: "AsyncRetrying | None" = None,
    retry_on_device_busy: bool = True,
    retry_on_device_failure: bool = False,
    **connection_kwargs: Any,
) -> AsyncModbusClient:
    """Create an asynchronous TCP Modbus client with automatic reconnect and request retry functionality.

    Args:
        host: The IP address or hostname of the Modbus server.
        port: The port number of the Modbus server (default is 502).
        unit_id: The unit ID to use for requests.
        timeout: Timeout in seconds, default 10.0s
        connect_timeout: Timeout for establishing connection, default 10.0s
        wait_between_requests: Wait time between requests in seconds (default: 0.0s)
        wait_after_connect: Wait time after connection establishment in seconds (default: 0.0s)
        auto_reconnect: Whether to automatically reconnect on connection loss (default: True).
                        Can be a custom AsyncRetrying instance when more control is needed.
        on_reconnected: Callback to be called after a successful reconnection.
        response_retry_strategy: Retry strategy for handling failed requests (default: None).
        retry_on_device_busy: Whether to retry on device busy errors (default: True).
                              Can be a custom AsyncRetrying instance when more control is needed.
        retry_on_device_failure: Whether to retry on device failure errors (default: False).
                                 Can be a custom AsyncRetrying instance when more control is needed.
        connection_kwargs: Additional connection parameters passed to `asyncio.open_connection` (e.g., SSL context)

    Returns:
        An instance of AsyncModbusClient configured for TCP transport.

    """
    smart_transport = AsyncSmartTransport(
        AsyncTcpTransport(
            host,
            port,
            timeout=timeout,
            connect_timeout=connect_timeout,
            **connection_kwargs,
        ),
        wait_between_requests=wait_between_requests,
        wait_after_connect=wait_after_connect,
        auto_reconnect=auto_reconnect,
        on_reconnected=on_reconnected,
        response_retry_strategy=response_retry_strategy,
        retry_on_device_busy=retry_on_device_busy,
        retry_on_device_failure=retry_on_device_failure,
    )
    return AsyncModbusClient(smart_transport, unit_id=unit_id)


def create_async_rtu_client(  # noqa: PLR0913
    port: str,
    *,
    unit_id: int,
    wait_between_requests: float = 0.0,
    wait_after_connect: float = 0.0,
    auto_reconnect: "bool | AsyncRetrying" = True,
    on_reconnected: Callable[[], Awaitable[None] | None] | None = None,
    response_retry_strategy: "AsyncRetrying | None" = None,
    retry_on_device_busy: bool = True,
    retry_on_device_failure: bool = False,
    **pyserial_options: Unpack[PySerialOptions],
) -> AsyncModbusClient:
    """Create an asynchronous RTU Modbus client with automatic reconnect and request retry functionality.

    Args:
        port: The port number of the Modbus server (default is 502).
        unit_id: The unit ID to use for requests.
        timeout: Timeout in seconds, default 10.0s
        connect_timeout: Timeout for establishing connection, default 10.0s
        wait_between_requests: Wait time between requests in seconds (default: 0.0s)
        wait_after_connect: Wait time after connection establishment in seconds (default: 0.0s)
        auto_reconnect: Whether to automatically reconnect on connection loss (default: True).
                        Can be a custom AsyncRetrying instance when more control is needed.
        on_reconnected: Callback to be called after a successful reconnection.
        response_retry_strategy: Retry strategy for handling failed requests (default: None).
        retry_on_device_busy: Whether to retry on device busy errors (default: True).
                              Can be a custom AsyncRetrying instance when more control is needed.
        retry_on_device_failure: Whether to retry on device failure errors (default: False).
                                 Can be a custom AsyncRetrying instance when more control is needed.
        pyserial_options: Additional connection parameters passed to `pyserial` (e.g., SSL context)

    Returns:
        An instance of AsyncModbusClient configured for TCP transport.

    """
    smart_transport = AsyncSmartTransport(
        AsyncRtuTransport(
            port,
            **pyserial_options,
        ),
        wait_between_requests=wait_between_requests,
        wait_after_connect=wait_after_connect,
        auto_reconnect=auto_reconnect,
        on_reconnected=on_reconnected,
        response_retry_strategy=response_retry_strategy,
        retry_on_device_busy=retry_on_device_busy,
        retry_on_device_failure=retry_on_device_failure,
    )
    return AsyncModbusClient(smart_transport, unit_id=unit_id)


__all__ = [
    "AsyncModbusClient",
    "AsyncRtuTransport",
    "AsyncSmartTransport",
    "AsyncTcpTransport",
    "create_async_rtu_client",
    "create_async_tcp_client",
]



"""Tests for tmodbus/utils/crc.py ."""

from tmodbus.utils.crc import calculate_crc16, validate_crc16


def test_calculate_crc() -> None:
    """Test combined compute/check CRC."""
    data = b"\x12\x34\x23\x45\x34\x56\x45\x67"
    assert calculate_crc16(data) == bytearray.fromhex("E2 DB")


def test_validate_crc() -> None:
    """Test framing with CRC."""
    assert validate_crc16(b"\x01") is False

    data = b"\x12\x34\x23\x45\x34\x56\x45\x67\xe2\xdb"
    assert validate_crc16(data)



"""Read Coils PDU Module."""

import struct
from typing import Self

from tmodbus.const import FunctionCode
from tmodbus.exceptions import InvalidRequestError, InvalidResponseError

from .base import BasePDU


class ReadCoilsPDU(BasePDU[list[bool]]):
    """Read Coils PDU."""

    function_code = FunctionCode.READ_COILS

    def __init__(self, start_address: int, quantity: int) -> None:
        """Initialize Read Coils PDU.

        Args:
            start_address: Starting address of the coils to read
            quantity: Number of coils to read

        Raises:
            ValueError: If start_address or quantity is invalid

        """
        if not (0 <= start_address < 65536):
            msg = "Address must be between 0 and 65535."
            raise ValueError(msg)
        self.start_address = start_address

        if not (1 <= quantity <= 2000):
            msg = "Quantity must be between 1 and 2000."
            raise ValueError(msg)
        self.quantity = quantity

    def encode_request(self) -> bytes:
        """Convert PDU to bytes.

        Returns:
            Bytes representation of the Read Coils PDU

        """
        return struct.pack(">BHH", self.function_code, self.start_address, self.quantity)

    def decode_response(self, response: bytes) -> list[bool]:
        """Decode the response PDU.

        Args:
            response: Response PDU bytes

        Returns:
            List of boolean values representing the coil states

        Raises:
            ValueError: If response format is invalid

        """
        # response format: function code + byte count + data
        try:
            function_code, byte_count = struct.unpack_from(">BB", response)
        except struct.error as e:
            msg = "Expected response to start with function code and byte count"
            raise InvalidResponseError(msg, response_bytes=response) from e

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:#04x}, received {function_code:#04x}"
            raise InvalidResponseError(msg, response_bytes=response)

        if len(response) != 2 + byte_count:
            msg = f"Invalid response PDU length: expected {2 + byte_count}, got {len(response)}"
            raise InvalidResponseError(msg, response_bytes=response)

        if byte_count != (self.quantity + 7) // 8:
            msg = f"Invalid byte count: expected {(self.quantity + 7) // 8}, got {byte_count}"
            raise InvalidResponseError(msg, response_bytes=response)

        coils: list[bool] = []
        for byte in response[2:]:
            coils.extend([bool(byte & (1 << bit)) for bit in range(8)])

        return coils[: self.quantity]  # Ensure we return only the requested quantity

    @classmethod
    def decode_request(cls, request: bytes) -> Self:
        """Decode Read Coils Request PDU.

        Args:
            request: the request bytes.

        Returns:
            ReadCoilsPDU: The decoded Read Coils Request PDU.

        """
        try:
            function_code, address, quantity = struct.unpack(">BHH", request)
        except struct.error as e:
            msg = "Expected request to start with function code, address, and quantity"
            raise InvalidRequestError(msg, request_bytes=request) from e

        if function_code != cls.function_code:
            msg = f"Invalid function code: expected {cls.function_code:#04x}, received {function_code:#04x}"
            raise InvalidRequestError(msg, request_bytes=request)

        return cls(address, quantity)

    def encode_response(self, value: list[bool]) -> bytes:
        """Convert PDU to bytes.

        Args:
            value: List of boolean values representing the coil states

        Returns:
            Bytes representation of the Read Coils Response PDU

        """
        byte_count = (len(value) + 7) // 8
        data = bytearray(byte_count)
        for i, v in enumerate(value):
            if v:
                data[i // 8] |= 1 << (i % 8)

        return (
            struct.pack(
                ">BB",
                self.function_code,
                byte_count,
            )
            + data
        )


class WriteSingleCoilPDU(BasePDU[bool]):
    """Write Single Coil PDU."""

    function_code = FunctionCode.WRITE_SINGLE_COIL
    rtu_response_data_length = 4  # address (2) + value (2)

    def __init__(
        self,
        address: int,
        value: bool,  # noqa: FBT001
    ) -> None:
        """Initialize Write Single Coil PDU.

        Args:
            address: Address of the coil to write
            value: Value to write (True for ON, False for OFF)

        Raises:
            ValueError: If address is invalid

        """
        super().__init__()
        self.address = address
        self.value = value

    def encode_request(self) -> bytes:
        """Convert PDU to bytes.

        Returns:
            Bytes representation of the Write Single Coil PDU

        """
        coil_value = 0xFF00 if self.value else 0x0000
        return struct.pack(">BHH", self.function_code, self.address, coil_value)

    def decode_response(self, response: bytes) -> bool:
        """Decode the response PDU.

        Args:
            response: Response PDU bytes

        Raises:
            InvalidResponseError: If response format is invalid

        """
        if response != self.encode_request():
            msg = "Expected response to match request"
            raise InvalidResponseError(msg, response_bytes=response)
        return self.value

    @classmethod
    def decode_request(cls, request: bytes) -> Self:
        """Decode Write Single Coil Request PDU.

        Args:
            request: the request bytes.

        Returns:
            WriteSingleCoilPDU: The decoded Write Single Coil Request PDU.

        """
        try:
            function_code, address, coil_value = struct.unpack(">BHH", request)
        except struct.error as e:
            msg = "Expected request to start with function code, address, and value"
            raise InvalidRequestError(msg, request_bytes=request) from e

        if function_code != cls.function_code:
            msg = f"Invalid function code: expected {cls.function_code:#04x}, received {function_code:#04x}"
            raise InvalidRequestError(msg, request_bytes=request)

        if coil_value not in (0x0000, 0xFF00):
            msg = f"Invalid coil value: {coil_value:#06x}"
            raise InvalidRequestError(msg, request_bytes=request)

        return cls(address, coil_value == 0xFF00)

    def encode_response(self, value: bool) -> bytes:  # noqa: FBT001
        """Encode the response PDU.

        Returns:
            Bytes representation of the Write Single Coil response PDU.

        Notes:
            For Write Single Coil, the response echoes the request.

        """
        coil_value = 0xFF00 if value else 0x0000
        return struct.pack(">BHH", self.function_code, self.address, coil_value)


class WriteMultipleCoilsPDU(BasePDU[int]):
    """Write Multiple Coils PDU."""

    function_code = FunctionCode.WRITE_MULTIPLE_COILS
    rtu_response_data_length = 4  # address (2) + quantity (2)

    def __init__(self, start_address: int, values: list[bool]) -> None:
        """Initialize Write Multiple Coils PDU.

        Args:
            start_address: Starting address of the coils to write
            values: List of boolean values representing the coil states

        Raises:
            ValueError: If start_address or values are invalid

        """
        if not (0 <= start_address < 65536):
            msg = "Address must be between 0 and 65535."
            raise ValueError(msg)

        self.address = start_address
        if not (1 <= len(values) <= 0x07B0):  # 1968 coils max
            msg = "Number of coils must be between 1 and 1968."
            raise ValueError(msg)

        self.values = values

    def encode_request(self) -> bytes:
        """Convert PDU to bytes.

        Returns:
            Bytes representation of the Write Multiple Coils PDU

        """
        # Calculate required byte count
        byte_count = (len(self.values) + 7) // 8

        # Prepare data bytes
        data = bytearray(byte_count)
        for i, value in enumerate(self.values):
            if value:
                data[i // 8] |= 1 << (i % 8)

        return (
            struct.pack(
                ">BHHB",
                self.function_code,
                self.address,
                len(self.values),
                byte_count,
            )
            + data
        )

    def decode_response(self, response: bytes) -> int:
        """Decode the response PDU.

        Args:
            response: Response PDU bytes

        Raises:
            InvalidResponseError: If response format is invalid

        """
        # Verify response: function code + starting address + quantity
        expected_response = struct.pack(
            ">BHH",
            self.function_code,
            self.address,
            len(self.values),
        )

        if response != expected_response:
            msg = "Device response does not match request"
            raise InvalidResponseError(msg, response_bytes=response)

        return len(self.values)

    @classmethod
    def decode_request(cls, request: bytes) -> Self:
        """Decode Write Multiple Coils Request PDU.

        Args:
            request: the request bytes.

        Returns:
            WriteMultipleCoilsPDU: The decoded Write Multiple Coils Request PDU.

        """
        # Expected format: FC (1) + start address (2) + quantity (2) + byte count (1) + data (N)
        if len(request) < 6:
            msg = "Request too short for Write Multiple Coils"
            raise InvalidRequestError(msg, request_bytes=request)

        function_code, start_address, quantity, byte_count = struct.unpack(">BHHB", request[:6])

        if function_code != cls.function_code:
            msg = f"Invalid function code: expected {cls.function_code:#04x}, received {function_code:#04x}"
            raise InvalidRequestError(msg, request_bytes=request)

        if not (1 <= quantity <= 0x07B0):
            msg = "Number of coils must be between 1 and 1968."
            raise InvalidRequestError(msg, request_bytes=request)

        expected_len = 6 + byte_count
        if len(request) != expected_len:
            msg = f"Invalid request length: expected {expected_len}, got {len(request)}"
            raise InvalidRequestError(msg, request_bytes=request)

        expected_byte_count = (quantity + 7) // 8
        if byte_count != expected_byte_count:
            msg = f"Invalid byte count: expected {expected_byte_count}, got {byte_count}"
            raise InvalidRequestError(msg, request_bytes=request)

        data = request[6:]
        values: list[bool] = []
        for byte in data:
            values.extend([(byte >> bit) & 1 == 1 for bit in range(8)])

        return cls(start_address, values[:quantity])

    def encode_response(self, value: int) -> bytes:
        """Encode the response PDU.

        Args:
            value: The number of coils written to.

        Returns:
            Bytes representation of the Write Multiple Coils response PDU.

        """
        return struct.pack(
            ">BHH",
            self.function_code,
            self.address,
            value,
        )



"""Async TCP Transport Layer Implementation.

Implements async Modbus TCP protocol transport based on asyncio, including MBAP header processing.
"""

import asyncio
import logging
import struct
from functools import partial
from typing import Any, TypeVar

from tmodbus.exceptions import (
    InvalidResponseError,
    ModbusConnectionError,
    ModbusResponseError,
    error_code_to_exception_map,
)
from tmodbus.pdu import BaseClientPDU
from tmodbus.utils.raw_traffic_logger import log_raw_traffic as base_log_raw_traffic

from .async_base import AsyncBaseTransport

RT = TypeVar("RT")

logger = logging.getLogger(__name__)
log_raw_traffic = partial(base_log_raw_traffic, "TCP")


class AsyncTcpTransport(AsyncBaseTransport):
    """Async Modbus TCP Transport Layer Implementation.

    Handles async Modbus TCP communication based on asyncio, including:
    - Async TCP socket connection management
    - MBAP header construction and parsing
    - Transaction identifier management
    - Async error handling and timeout management
    """

    _reader: asyncio.StreamReader | None = None
    _writer: asyncio.StreamWriter | None = None
    _next_transaction_id: int = 1

    _communication_lock = asyncio.Lock()  # Prevents concurrent access to the transport layer
    _last_request_finished_at: float = 0.0

    def __init__(
        self,
        host: str,
        port: int = 502,
        *,
        timeout: float = 10.0,
        connect_timeout: float = 10.0,
        **connection_kwargs: Any,
    ) -> None:
        """Initialize async TCP transport layer.

        Args:
            host: Target host IP address or domain name
            port: Target port, default 502 (Modbus TCP standard port)
            timeout: Timeout in seconds, default 10.0s
            connect_timeout: Timeout for establishing connection, default 10.0s
            connection_kwargs: Additional connection parameters passed to `asyncio.open_connection` (e.g., SSL context)

        Raises:
            ValueError: When parameters are invalid
            TypeError: When parameter types are incorrect

        """
        if not 0 < port < 65535:
            msg = "Port must be an integer between 1-65535."
            raise ValueError(msg)
        if timeout <= 0:
            msg = "Timeout must be a positive number."
            raise ValueError(msg)
        if connect_timeout <= 0:
            msg = "Connect timeout must be a positive number."
            raise ValueError(msg)

        self.host = host
        self.port = port
        self.timeout = timeout
        self.connect_timeout = connect_timeout
        self.connection_kwargs = connection_kwargs

    async def open(self) -> None:
        """Async establish TCP connection."""
        async with self._communication_lock:
            if self.is_open():
                logger.debug("Async TCP connection already open: %s:%d", self.host, self.port)
                return

            try:
                self._reader, self._writer = await asyncio.wait_for(
                    asyncio.open_connection(self.host, self.port, **self.connection_kwargs),
                    timeout=self.connect_timeout,
                )

                logger.info("Async TCP connection established: %s:%d", self.host, self.port)
            except TimeoutError:
                logger.warning("Async TCP connection timeout: %s:%d", self.host, self.port, exc_info=True)
                raise
            except Exception as e:
                logger.exception("Async TCP connection error: %s:%d", self.host, self.port)
                raise ModbusConnectionError from e

    async def close(self) -> None:
        """Close TCP connection."""
        async with self._communication_lock:
            if not self._writer:
                logger.debug("Async TCP connection already closed: %s:%d", self.host, self.port)
                return

            try:
                self._writer.close()
                await self._writer.wait_closed()
                logger.info("Async TCP connection closed: %s:%d", self.host, self.port)
            except Exception as e:  # noqa: BLE001
                logger.debug("Error during async connection close (ignorable): %s", e)
            finally:
                self._reader = None
                self._writer = None

    def is_open(self) -> bool:
        """Async check TCP connection status."""
        if self._writer is None or self._reader is None:
            return False

        return not self._writer.is_closing()

    def _get_transaction_id(self) -> int:
        """Get the next transaction ID."""
        current_transaction_id = self._next_transaction_id
        self._next_transaction_id = (self._next_transaction_id + 1) % 0x10000  # 16-bit wraparound
        return current_transaction_id

    async def send_and_receive(self, unit_id: int, pdu: BaseClientPDU[RT]) -> RT:
        """Async send PDU and receive response."""
        # Ensure that only one request is processed at a time
        async with self._communication_lock:
            return await self._do_send_and_receive(unit_id, pdu)

    async def _do_send_and_receive(self, unit_id: int, pdu: BaseClientPDU[RT]) -> RT:
        """Async send PDU and receive response.

        Implements complete async TCP protocol communication flow:
        1. Build MBAP header
        2. Async send request (MBAP header + PDU)
        3. Async receive response MBAP header
        4. Validate MBAP header
        5. Async receive response PDU
        6. Return response PDU
        """
        if not self.is_open():
            msg = "Not connected."
            raise ModbusConnectionError(msg)

        # 1. Generate transaction ID and build MBAP header
        current_transaction_id = self._get_transaction_id()

        request_pdu_bytes = pdu.encode_request()  # Convert PDU to bytes

        # MBAP header format:
        # - Transaction ID (2 bytes):  Transaction identifier
        # - Protocol ID (2 bytes): Protocol identifier, fixed to 0x0000
        # - Length (2 bytes): Length of following bytes (Unit ID + PDU)
        # - Unit ID (1 byte): Unit identifier (slave address)
        mbap_header = struct.pack(
            ">HHHB",  # Big endian: 2 shorts, 1 short, 1 byte
            current_transaction_id,  # Transaction ID
            0x0000,  # Protocol ID
            len(request_pdu_bytes) + 1,  # Length (PDU length + 1 byte for Unit ID)
            unit_id,  # Unit ID
        )

        # 2. Build complete request frame
        request_frame = mbap_header + request_pdu_bytes
        log_raw_traffic("sent", request_frame)

        # 3. Async send request
        if self._writer is None:
            msg = "Connection not established."
            raise ModbusConnectionError(msg)
        self._writer.write(request_frame)
        await asyncio.wait_for(self._writer.drain(), timeout=self.timeout)

        # 4. Async receive response MBAP header (7 bytes)
        response_mbap = await self._receive_exact(7)

        # 5. Parse response MBAP header
        (
            response_transaction_id,
            response_protocol_id,
            response_length,
            response_unit_id,
        ) = struct.unpack(">HHHB", response_mbap)

        # 6. Validate MBAP header
        if response_transaction_id != current_transaction_id:
            msg = (
                f" Transaction ID mismatch: expected {current_transaction_id:#04x}, "
                f"received {response_transaction_id:#04x}"
            )
            raise InvalidResponseError(msg, response_bytes=response_mbap)

        if response_protocol_id != 0x0000:
            msg = f"Invalid Protocol ID: expected 0x0000, received 0x{response_protocol_id:04x}"
            raise InvalidResponseError(msg, response_bytes=response_mbap)

        if response_unit_id != unit_id:
            msg = f"Unit ID mismatch: expected {unit_id:#04x}, received {response_unit_id:#04x}"
            raise InvalidResponseError(msg, response_bytes=response_mbap)

        # 7. Async receive response PDU
        pdu_length = response_length - 1  # Subtract 1 byte for Unit ID
        if pdu_length <= 0:
            msg = f"Invalid PDU length: {pdu_length}"
            raise InvalidResponseError(msg, response_bytes=response_mbap)

        try:
            response_pdu_bytes = await self._receive_exact(pdu_length)
        except ModbusConnectionError as e:
            log_raw_traffic("recv", e.response_bytes, is_error=True)
            raise
        else:
            log_raw_traffic("recv", response_mbap + response_pdu_bytes)

        # 8.Check if it's an exception response
        if len(response_pdu_bytes) > 0 and response_pdu_bytes[0] & 0x80:  # Exception response
            function_code = response_pdu_bytes[0] & 0x7F  # Remove exception flag bit
            exception_code = response_pdu_bytes[1] if len(response_pdu_bytes) > 1 else 0

            error_class = error_code_to_exception_map.get(exception_code, ModbusResponseError)

            raise error_class(exception_code, function_code)

        return pdu.decode_response(response_pdu_bytes)

    async def _receive_exact(self, length: int) -> bytes:
        """Async receive exact length of data.

        Args:
            length: Number of bytes to receive

        Returns:
            Received data

        Raises:
            TimeoutError: Receive timeout
            ModbusConnectionError: Connection error

        """
        if self._reader is None:
            msg = "Connection not established."
            raise ModbusConnectionError(msg)

        try:
            return await asyncio.wait_for(self._reader.readexactly(length), timeout=self.timeout)
        except asyncio.IncompleteReadError as e:
            msg = f"Received incomplete data: expected {length} bytes, got {len(e.partial)} bytes"
            raise ModbusConnectionError(msg, bytes_read=e.partial) from e
        except TimeoutError as e:
            msg = f"Receive timeout: expected {length} bytes, but timed out after {self.timeout} seconds"
            raise TimeoutError(msg) from e
        except Exception as e:
            msg = f"Failed to read {length} bytes"
            raise ModbusConnectionError(msg) from e
